#!/usr/bin/env python
# network_connector.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (network_connector.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: network_connector.

.. role:: red

.. raw:: html

    <a href="https://bitdust.io/automats/network_connector/network_connector.png" target="_blank">
    <img src="https://bitdust.io/automats/network_connector/network_connector.png" style="max-width:100%;">
    </a>

The ``network_connector()`` machine is needed to monitor status of the Internet connection.

It will periodically check for incoming traffic and start STUN discovery procedure
to detect connection status and possible external IP changes.

If BitDust get disconnected it will ping "Google dot com" (joke) to check what is going on.


EVENTS:
    * :red:`all-network-transports-disabled`
    * :red:`all-network-transports-ready`
    * :red:`check-reconnect`
    * :red:`connection-done`
    * :red:`gateway-is-not-started`
    * :red:`got-network-info`
    * :red:`init`
    * :red:`internet-failed`
    * :red:`internet-success`
    * :red:`network-down`
    * :red:`network-transport-state-changed`
    * :red:`network-transports-verified`
    * :red:`network-up`
    * :red:`reconnect`
    * :red:`timer-1hour`
    * :red:`timer-5sec`
    * :red:`upnp-done`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from six.moves import range

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 16

#------------------------------------------------------------------------------

import sys
import time

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in network_connector.py')

from twisted.internet import threads

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat
from bitdust.automats import global_state

from bitdust.lib import net_misc
from bitdust.lib import misc
from bitdust.lib import strng

from bitdust.services import driver

from bitdust.main import settings
from bitdust.main import shutdowner

from bitdust.p2p import p2p_stats

#------------------------------------------------------------------------------

_NetworkConnector = None
_CounterSuccessConnections = 0
_CounterFailedConnections = 0
_LastSuccessConnectionTime = 0

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _NetworkConnector
    if event is None:
        return _NetworkConnector
    if _NetworkConnector is None:
        _NetworkConnector = NetworkConnector(
            name='network_connector',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is not None:
        _NetworkConnector.automat(event, *args, **kwargs)
    return _NetworkConnector


def Destroy():
    """
    Destroy network_connector() automat and remove its instance from memory.
    """
    global _NetworkConnector
    if _NetworkConnector is None:
        return
    _NetworkConnector.destroy()
    del _NetworkConnector
    _NetworkConnector = None


class NetworkConnector(automat.Automat):
    """
    Class to monitor Internet connection and reconnect when needed.
    """

    fast = False

    timers = {
        'timer-1hour': (3600, ['DISCONNECTED']),
        'timer-5sec': (30.0, ['DISCONNECTED', 'CONNECTED']),
    }

    managed_services = [
        'service_udp_datagrams',
        'service_service_entangled_dht',
        'service_ip_port_responder',
        'service_my_ip_port',
        'service_private_messages',
    ]

    def init(self):
        self.last_upnp_time = 0
        self.last_reconnect_time = 0
        self.last_internet_state = 'disconnected'
        self.last_bytes_in_counter = 0
        net_misc.SetConnectionDoneCallbackFunc(ConnectionDoneCallback)
        net_misc.SetConnectionFailedCallbackFunc(ConnectionFailedCallback)

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        global_state.set_global_state('NETWORK ' + newstate)
        if driver.is_on('service_p2p_hookups'):
            from bitdust.p2p import p2p_connector
            from bitdust.system import tray_icon
            p2p_connector.A('network_connector.state', newstate)
            tray_icon.state_changed(self.state, p2p_connector.A().state)
        if oldstate != 'CONNECTED' and newstate == 'CONNECTED':
            # TODO: redesign the state machine to cover that
            if self.last_bytes_in_counter < p2p_stats.get_total_bytes_in():
                lg.info('HELLO BITDUST WORLD!')
            else:
                lg.warn('no incoming traffic detected yet, potentially a signal that my node is offline')
            self.last_bytes_in_counter = p2p_stats.get_total_bytes_in()

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'START_UP'
                self.Disconnects = 0
                self.Reset = False
                self.ColdStart = True
                self.doCheckNetworkInterfaces(*args, **kwargs)
        #---CONNECTED---
        elif self.state == 'CONNECTED':
            if event == 'reconnect' or (event == 'timer-5sec' and (self.Reset or not self.isConnectionAlive(*args, **kwargs))):
                self.state = 'DOWN'
                self.Disconnects = 0
                self.Reset = False
                self.doSetDown(*args, **kwargs)
            elif event == 'check-reconnect':
                self.state = 'TRANSPORTS?'
                self.doVerifyTransports(*args, **kwargs)
        #---NETWORK?---
        elif self.state == 'NETWORK?':
            if event == 'got-network-info' and not self.isNetworkActive(*args, **kwargs):
                self.state = 'DISCONNECTED'
            elif event == 'got-network-info' and self.isNetworkActive(*args, **kwargs) and self.isCurrentInterfaceActive(*args, **kwargs):
                self.state = 'INTERNET?'
                self.doPingGoogleDotCom(*args, **kwargs)
            elif event == 'got-network-info' and self.isNetworkActive(*args, **kwargs) and not self.isCurrentInterfaceActive(*args, **kwargs):
                self.state = 'UP'
                self.doSetUp(*args, **kwargs)
        #---INTERNET?---
        elif self.state == 'INTERNET?':
            if event == 'internet-failed':
                self.state = 'DISCONNECTED'
            elif event == 'internet-success':
                self.state = 'UP'
                self.doSetUp(*args, **kwargs)
        #---DISCONNECTED---
        elif self.state == 'DISCONNECTED':
            if event == 'reconnect' or event == 'check-reconnect' or event == 'timer-1hour' or (event == 'timer-5sec' and (self.Disconnects < 3 or self.Reset)) or (event == 'connection-done' and self.isTimePassed(*args, **kwargs)):
                self.state = 'DOWN'
                self.doRememberTime(*args, **kwargs)
                self.Disconnects += 1
                self.Reset = False
                self.doSetDown(*args, **kwargs)
        #---UP---
        elif self.state == 'UP':
            if not self.ColdStart and event == 'network-up' and not self.isNeedUPNP(*args, **kwargs):
                self.state = 'TRANSPORTS?'
                self.doStartNetworkTransports(*args, **kwargs)
            elif event == 'reconnect' or event == 'check-reconnect':
                self.Reset = True
            elif self.ColdStart and event == 'network-up':
                self.state = 'TRANSPORTS?'
                self.doColdStartNetworkTransports(*args, **kwargs)
                self.ColdStart = False
            elif not self.ColdStart and event == 'network-up' and self.isNeedUPNP(*args, **kwargs):
                self.state = 'UPNP'
                self.doUPNP(*args, **kwargs)
        #---DOWN---
        elif self.state == 'DOWN':
            if event == 'network-down':
                self.state = 'NETWORK?'
                self.doCheckNetworkInterfaces(*args, **kwargs)
        #---TRANSPORTS?---
        elif self.state == 'TRANSPORTS?':
            if event == 'reconnect' or event == 'check-reconnect':
                self.Reset = True
            elif not self.Reset and (
                (event == 'all-network-transports-ready' or event == 'network-transports-verified' or event == 'network-transport-state-changed') and (self.isAllReady(*args, **kwargs) and self.isAllListening(*args, **kwargs))
            ):
                self.state = 'CONNECTED'
            elif self.Reset and ((event == 'all-network-transports-ready' or event == 'network-transports-verified' or event == 'network-transport-state-changed') and self.isAllReady(*args, **kwargs)):
                self.state = 'DOWN'
                self.Reset = False
                self.Disconnects = 0
                self.doSetDown(*args, **kwargs)
            elif (event == 'all-network-transports-disabled' or event == 'gateway-is-not-started' or event == 'network-transport-state-changed') and (self.isAllReady(*args, **kwargs) and not self.isAllListening(*args, **kwargs)):
                self.state = 'DISCONNECTED'
        #---START_UP---
        elif self.state == 'START_UP':
            if event == 'got-network-info' and not self.isNetworkActive(*args, **kwargs):
                self.state = 'DISCONNECTED'
                self.Disconnects = 3
            elif event == 'reconnect' or event == 'check-reconnect':
                self.state = 'UP'
                self.doSetUp(*args, **kwargs)
        #---UPNP---
        elif self.state == 'UPNP':
            if event == 'upnp-done':
                self.state = 'TRANSPORTS?'
                self.doStartNetworkTransports(*args, **kwargs)
            elif event == 'reconnect' or event == 'check-reconnect':
                self.Reset = True
        return None

    def isNeedUPNP(self, *args, **kwargs):
        if not settings.enableUPNP():
            return False
        if driver.is_on('service_tcp_transport'):
            try:
                from bitdust.transport.tcp import tcp_node
                if int(tcp_node.get_internal_port()) != int(settings.getTCPPort()):
                    return True
            except:
                lg.exc()
                return False
        return time.time() - self.last_upnp_time > 60*60

    def isConnectionAlive(self, *args, **kwargs):
        # miss = 0
        if driver.is_on('service_udp_datagrams'):
            from bitdust.lib import udp
            if time.time() - udp.get_last_datagram_time() < 60:
                if settings.enableUDP() and settings.enableUDPreceiving():
                    return True
        # else:
        #     miss += 1
        if driver.is_on('service_gateway'):
            from bitdust.transport import gateway
            if time.time() - gateway.last_inbox_time() < 60:
                return True
            transport_states = [t.state for t in list(gateway.transports().values())]
            if 'LISTENING' in transport_states:
                return True
            if 'STARTING' in transport_states:
                return True
        # else:
        #     miss += 1
        # if miss >= 2:
        #     return True
        # return False ... still testing that approach
        return True

    def isNetworkActive(self, *args, **kwargs):
        ips = args[0]
        if _Debug:
            lg.args(_DebugLevel, ips=ips)
        return len(ips) > 0

    def isCurrentInterfaceActive(self, *args, **kwargs):
        # TODO: not sure about external IP
        # because if you have a white IP it should be the same as your local IP
        loc = strng.to_bin(misc.readLocalIP())
        ext = strng.to_bin(misc.readExternalIP())
        ips = args[0]
        if _Debug:
            lg.args(_DebugLevel, local=loc, external=ext)
        return (loc in ips) or (ext in ips)

    def isTimePassed(self, *args, **kwargs):
        return time.time() - self.last_reconnect_time < 15

    def isAllListening(self, *args, **kwargs):
        """
        Condition method.
        """
        if not driver.is_on('service_gateway'):
            if _Debug:
                lg.out(_DebugLevel, 'network_connector.isAllListening returning False : service_gateway is OFF')
            return False
        from bitdust.transport import gateway
        transports = list(gateway.transports().values())
        for t in transports:
            if t.state != 'LISTENING':
                if _Debug:
                    lg.out(_DebugLevel, 'network_connector.isAllListening returning False : transport %s is not LISTENING' % t)
                return False
        if _Debug:
            lg.out(_DebugLevel, 'network_connector.isAllListening returning True')
        return True

    def isAllReady(self, *args, **kwargs):
        """
        Condition method.
        """
        if not driver.is_on('service_gateway'):
            if _Debug:
                lg.out(_DebugLevel, 'network_connector.isAllReady returning False : service_gateway is OFF')
            return False
        LISTENING_count = 0
        OFFLINE_count = 0
        from bitdust.transport import gateway
        transports = list(gateway.transports().values())
        for t in transports:
            if t.state != 'OFFLINE' and t.state != 'LISTENING':
                if _Debug:
                    lg.out(_DebugLevel, 'network_connector.isAllReady returning False : transport %s is not READY yet' % t)
                return False
            if t.state == 'OFFLINE':
                OFFLINE_count += 1
            if t.state == 'LISTENING':
                LISTENING_count += 1
        if _Debug:
            lg.out(_DebugLevel, 'network_connector.isAllReady returning True : all transports READY')
            lg.out(_DebugLevel, '    OFFLINE transports:%d, LISTENING transports: %d' % (OFFLINE_count, LISTENING_count))
        return True

    def doSetUp(self, *args, **kwargs):
        for svc_name in self.managed_services:
            if driver.is_on(svc_name):
                if driver.is_suspended(svc_name):
                    if not driver.resume(svc_name):
                        lg.err('resuming network service %r failed' % svc_name)
                    else:
                        lg.info('network service resumed: %r' % svc_name)
                else:
                    lg.warn('network service %r was not suspended' % svc_name)
            else:
                lg.warn('network service %r was not ON' % svc_name)
        self.automat('network-up')

    def doSetDown(self, *args, **kwargs):
        for svc_name in [
            'service_gateway',
        ] + list(reversed(self.managed_services)):
            if driver.is_on(svc_name):
                if not driver.is_suspended(svc_name):
                    if not driver.suspend(svc_name):
                        lg.err('suspending network service %r failed' % svc_name)
                    else:
                        lg.info('network service suspended: %r' % svc_name)
                else:
                    lg.warn('network service %r already suspended' % svc_name)
            else:
                lg.warn('network service %r was not ON' % svc_name)
        self.automat('network-down')

    def doUPNP(self, *args, **kwargs):
        self.last_upnp_time = time.time()
        UpdateUPNP()

    def doPingGoogleDotCom(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'network_connector.doPingGoogleDotCom')
        net_misc.TestInternetConnection().addCallbacks(lambda x: self.automat('internet-success', 'connected'), lambda x: self.automat('internet-failed', 'disconnected'))

    def doCheckNetworkInterfaces(self, *args, **kwargs):
        start_time = time.time()
        #         if bpio.Linux():
        #             def _call():
        #                 return net_misc.getNetworkInterfaces()
        #
        #             def _done(result, start_time):
        #                 if _Debug:
        #                     lg.out(_DebugLevel, 'network_connector.doCheckNetworkInterfaces DONE: %s in %d seconds' % (str(result), time.time() - start_time))
        #                 if not result:
        #                     lg.err('no network interfaces found')
        #                 self.automat('got-network-info', result)
        #             d = threads.deferToThread(_call)
        #             d.addBoth(_done, start_time)
        #         else:
        ips = net_misc.getNetworkInterfaces()
        if _Debug:
            lg.args(_DebugLevel, result=ips, tm=(time.time() - start_time))
        if not ips:
            lg.err('no network interfaces found')
        self.automat('got-network-info', ips)

    def doRememberTime(self, *args, **kwargs):
        self.last_reconnect_time = time.time()

    def doStartNetworkTransports(self, *args, **kwargs):
        """
        Action method.
        """
        if not driver.is_on('service_gateway'):
            self.automat('gateway-is-not-started')
            return
        if not driver.is_suspended('service_gateway'):
            self.automat('all-network-transports-ready')
            return
        restarted_transports = driver.resume('service_gateway')
        if len(restarted_transports) == 0:
            self.automat('all-network-transports-ready')

    def doColdStartNetworkTransports(self, *args, **kwargs):
        """
        Action method.
        """
        if not driver.is_on('service_gateway'):
            self.automat('gateway-is-not-started')
            return
        if not driver.is_suspended('service_gateway'):
            self.automat('all-network-transports-ready')
            return
        restarted_transports = driver.resume('service_gateway', cold_start=True)
        if len(restarted_transports) == 0:
            self.automat('all-network-transports-ready')

    def doVerifyTransports(self, *args, **kwargs):
        """
        Action method.
        """
        if not driver.is_on('service_gateway'):
            self.automat('gateway-is-not-started')
            return
        from bitdust.transport import gateway

        def _transports_verified(all_results):
            if _Debug:
                lg.out(_DebugLevel, 'network_connector._transports_verified : %s' % str(all_results))
            order, all_results = all_results
            not_valid_count = 0
            restarts_count = 0
            if len(order) == 0:
                self.automat('network-transports-verified')
                return
            for proto in order:
                if not all_results[proto]:
                    not_valid_count += 1
            for priority in range(len(order)):
                proto = order[priority]
                if not all_results[proto]:
                    if _Debug:
                        lg.out(_DebugLevel, '    [%s] at position %d needs restart' % (proto, priority))
                    gateway.transport(proto).automat('restart')
                    restarts_count += 1
                    if not_valid_count > 1:  # this one failed, 2 other failed as well
                        self.automat('network-transports-verified')
                        return
                    continue
                if not_valid_count > 0:
                    if _Debug:
                        lg.out(_DebugLevel, '    skip %d transport [%s]' % (priority, proto))
                    if restarts_count == 0:
                        if _Debug:
                            lg.out(_DebugLevel, '    but no restarts and %d:[%s] is valid' % (priority, proto))
                        self.automat('network-transports-verified')
                    return
                if _Debug:
                    lg.out(_DebugLevel, '        [%s] at position %d is fine, skip other transports' % (proto, priority))
                self.automat('network-transports-verified')
                return

        gateway.verify().addCallback(_transports_verified)


#------------------------------------------------------------------------------


def UpdateUPNP():
    """
    Use ``lib.run_upnpc`` to configure UPnP device to create a port forwarding.
    """
    if _Debug:
        lg.out(_DebugLevel, 'network_connector.UpdateUPNP ')
    protos_need_upnp = set([
        'tcp',
    ])
    if not settings.enableTCP():
        # need to update only enabled protocols
        protos_need_upnp.discard('tcp')

    def _update_next_proto():
        if len(protos_need_upnp) == 0:
            lg.out(_DebugLevel, 'network_connector.update_upnp done, sending "upnp-done" event')
            A('upnp-done')
            return
        if _Debug:
            lg.out(_DebugLevel, 'network_connector.UpdateUPNP._update_next_proto ' + str(protos_need_upnp))
        proto = protos_need_upnp.pop()
        protos_need_upnp.add(proto)
        port = -1
        if proto == 'tcp':
            port = settings.getTCPPort()
        if port > 0:
            d = threads.deferToThread(_call_upnp, port)
            d.addCallback(_upnp_proto_done, proto)
        else:
            reactor.callLater(0, _upnp_proto_done, proto)  # @UndefinedVariable

    def _call_upnp(port):
        # start messing with upnp settings
        # success can be false if you're behind a router that doesn't support upnp
        # or if you are not behind a router at all and have an external ip address
        from bitdust.system import run_upnpc
        shutdowner.A('block')
        success, port = run_upnpc.update(port)
        shutdowner.A('unblock')
        return (success, port)

    def _upnp_proto_done(result, proto):
        if _Debug:
            lg.out(_DebugLevel, 'network_connector.UpdateUPNP._upnp_proto_done %s: %s' % (proto, str(result)))
        if result[0] == 'upnp-done':
            if proto == 'tcp':
                if str(settings.getTCPPort()) != str(result[1]).strip():
                    lg.out(_DebugLevel, '    !!!!!!!!!! created a new port mapping, TCP port were changed: %s -> %s' % (settings.getTCPPort(), str(result[1])))
                settings.setTCPPort(result[1])
        protos_need_upnp.discard(proto)
        reactor.callLater(0, _update_next_proto)  # @UndefinedVariable

    _update_next_proto()


#------------------------------------------------------------------------------


def ConnectionDoneCallback(param, proto, info):
    global _CounterSuccessConnections
    global _LastSuccessConnectionTime
    _CounterSuccessConnections += 1
    _LastSuccessConnectionTime = time.time()
    A('connection-done')


def ConnectionFailedCallback(param, proto, info):
    global _CounterFailedConnections
    if proto is not 'udp' and proto is not 'proxy':
        _CounterFailedConnections += 1
    A('connection-failed')



#!/usr/bin/python
# packetid.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (packetid.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: packetid.

Various methods to work with packet/path/backup ID.

We have a standard way of making the PacketID strings for many packets.
Note - we return strings and not integers.

Packet ID consists of several parts:
    <path ID>/<version Name>/<block number>-<supplier number>-<'Data'|'Parity'>

Backup ID is just a short form:
    <path ID>/<version Name>

Remote ID may have a different forms:
    - full:     alice@idhost.org:0/0/1/0/F20131120053803PM/0-1-Data
    - backupID: alice@idhost.org:0/0/1/0/F20131120053803PM
    - pathID:   alice@idhost.org:0/0/1/0

Here is:
    - customer:      alice@idhost.org
    - pathID:        0/0/1/0
    - versionName:   F20131120053803PM
    - blockNum :     0
    - supplierNum :  1
    - dataORparity : Data


See module ``p2p.backup_fs`` to learn how Path ID is generated from file or folder path.
"""

from __future__ import absolute_import
import time
import re
import random

#------------------------------------------------------------------------------

_LastUniqueNumber = 0

#------------------------------------------------------------------------------


def UniqueID():
    """
    Generate a unique string ID.

    We wrap around every billion, old packets should be gone by then
    """
    global _LastUniqueNumber
    if _LastUniqueNumber == 0:
        _LastUniqueNumber = int(time.time()*100.0)
    _LastUniqueNumber += 1
    if _LastUniqueNumber > 100000000000000:
        _LastUniqueNumber = int(time.time()*100.0)
    return str(int(random.getrandbits(8))) + str(_LastUniqueNumber)


def MakePacketID(backupID, blockNumber, supplierNumber, dataORparity, normalize_key_alias=True):
    """
    Create a full packet ID from backup ID and other parts
    Such call:

        MakePacketID('alice@idhost.org:0/0/1/0/F20131120053803PM', 1234, 63, 'Data')

    will return:

        'master$alice@idhost.org:0/0/1/0/F20131120053803PM/1234-63-Data'
    """
    if normalize_key_alias:
        if '$' not in backupID:
            backupID = 'master$' + backupID
    return '%s/%d-%d-%s' % (
        backupID,
        blockNumber,
        supplierNumber,
        dataORparity,
    )


def MakeBackupID(customer=None, path_id='0', version=None, normalize_key_alias=True, key_alias=None):
    """
    Run:

        MakeBackupID('alice@idhost.org', '0/0/1/0', 'F20131120053803PM')

    Will create a string like that:

        "master$alice@idhost.org:0/0/1/0/F20131120053803PM"
    """
    if normalize_key_alias:
        if not customer:
            from bitdust.userid import my_id
            customer = my_id.getGlobalID(key_alias=key_alias or 'master')
    if customer:
        if '$' not in customer and normalize_key_alias:
            customer = '{}${}'.format(key_alias or 'master', customer)
        if version:
            return '{}:{}/{}'.format(customer, path_id, version)
        return '{}:{}'.format(customer, path_id)
    if version:
        return '{}/{}'.format(path_id, version)
    return path_id


def Valid(packetID):
    """
    Must be in short form, without global user ID:

        0/1/2/F20131120053803PM/3-4-Data
        0/1/2/F20131120053803PM
        0/1/2

    """
    user, _, shortPacketID = packetID.rpartition(':')
    if user:
        return False
    head, x, tail = shortPacketID.rpartition('/')
    pathID = ''
    version = ''
    if not x and not head:
        # this seems to be a shortest pathID: 0, 1, 2, ...
        try:
            int(tail)
        except:
            return False
        pathID = tail
    if tail.endswith('-Data') or tail.endswith('-Parity'):
        if not IsPacketNameCorrect(tail):
            return False
        pathID, _, version = head.rpartition('/')
    else:
        pathID = shortPacketID
        if IsCanonicalVersion(tail):
            version = tail
            pathID = head
    if version and not IsCanonicalVersion(version):
        return False
    if not IsPathIDCorrect(pathID):
        return False
    return True


def Split(packetID, normalize_key_alias=True):
    """
    Split a full packet ID into tuple of 5 parts:

        packetid.Split("alice@idhost.org:0/0/1/0/F20131120053803PM/0-1-Data")
        ('master$alice@idhost.org', '0/0/1/0/F20131120053803PM', 0, 1, 'Data')
    """
    try:
        backupID, _, fileName = packetID.rpartition('/')
        blockNum, supplierNum, dataORparity = fileName.split('-')
        blockNum = int(blockNum)
        supplierNum = int(supplierNum)
        customerGlobalID, _, remotePathWithVersion = backupID.rpartition(':')
    except:
        return None, None, None, None, None
    if normalize_key_alias:
        if '$' not in customerGlobalID:
            customerGlobalID = 'master$' + customerGlobalID
    return customerGlobalID, remotePathWithVersion, blockNum, supplierNum, dataORparity


def SplitFull(packetID, normalize_key_alias=True):
    """
    Almost the same but return 6 parts:

        packetid.SplitFull("alice@idhost.org:0/0/1/0/F20131120053803PM/0-1-Data")
        ('master$alice@idhost.org', '0/0/1/0', 'F20131120053803PM', 0, 1, 'Data')
    """
    try:
        backupID, _, fileName = packetID.rpartition('/')
        pathID, _, versionName = backupID.rpartition('/')
        blockNum, supplierNum, dataORparity = fileName.split('-')
        blockNum = int(blockNum)
        supplierNum = int(supplierNum)
        customerGlobalID, _, remotePath = pathID.rpartition(':')
    except:
        return None, None, None, None, None, None
    if normalize_key_alias:
        if '$' not in customerGlobalID:
            customerGlobalID = 'master$' + customerGlobalID
    return customerGlobalID, remotePath, versionName, blockNum, supplierNum, dataORparity


def SplitVersionFilename(packetID, normalize_key_alias=True):
    """
    Return 4 parts:

        packetid.SplitVersionFilename("alice@idhost.org:0/0/1/0/F20131120053803PM/0-1-Data")
        ('master$alice@idhost.org', '0/0/1/0', 'F20131120053803PM', '0-1-Data')
    """
    try:
        backupID, _, fileName = packetID.rpartition('/')
        pathID, _, versionName = backupID.rpartition('/')
        customerGlobalID, _, remotePath = pathID.rpartition(':')
    except:
        return None, None, None, None
    if normalize_key_alias:
        if '$' not in customerGlobalID:
            customerGlobalID = 'master$' + customerGlobalID
    return customerGlobalID, remotePath, versionName, fileName


def SplitBackupID(backupID, normalize_key_alias=True):
    """
    This takes a backup ID string and split by 3 parts:

        packetid.SplitBackupID('alice@idhost.org:0/0/1/0/F20131120053803PM')
        ('master$alice@idhost.org', '0/0/1/0', 'F20131120053803PM')
    """
    try:
        pathID, _, versionName = backupID.rpartition('/')
        customerGlobalID, _, pathID = pathID.rpartition(':')
    except:
        return None, None, None
    if normalize_key_alias:
        if '$' not in customerGlobalID:
            customerGlobalID = 'master$' + customerGlobalID
    return customerGlobalID, pathID, versionName


def SplitBackupIDFull(backupID, normalize_key_alias=True):
    """
    This takes a backup ID string and split by 4 parts:

        packetid.SplitBackupID('share_1234$alice@idhost.org:0/0/1/0/F20131120053803PM')
        ('share_1234', 'alice@idhost.org', '0/0/1/0', 'F20131120053803PM')
    """
    try:
        pathID, _, versionName = backupID.rpartition('/')
        customerGlobalID, _, pathID = pathID.rpartition(':')
        keyAlias, _, customerGlobalID = customerGlobalID.partition('$')
    except:
        return None, None, None, None
    if normalize_key_alias:
        if not keyAlias:
            keyAlias = 'master'
    return keyAlias, customerGlobalID, pathID, versionName


def SplitPacketID(packetID, normalize_key_alias=True):
    """
    This takes a full packet ID string and split by 2 parts:

        packetid.SplitBackupID('alice@idhost.org:0/0/1/0/F20131120053803PM/1-2-Data')
        ('master$alice@idhost.org', '0/0/1/0/F20131120053803PM/1-2-Data')
    """
    try:
        customerGlobalID, _, dataID = packetID.rpartition(':')
    except:
        return None, None
    if normalize_key_alias:
        if '$' not in customerGlobalID:
            customerGlobalID = 'master$' + customerGlobalID
    return customerGlobalID, dataID


def SplitKeyOwnerData(packetID):
    """
    This takes a full packet ID string and split by 3 parts:

        packetid.SplitBackupID('abcd$alice@idhost.org:0/0/1/0/F20131120053803PM/1-2-Data')
        ('abcd', 'alice@idhost.org', '0/0/1/0/F20131120053803PM/1-2-Data')
    """
    try:
        fullGlobalID, _, dataID = packetID.rpartition(':')
    except:
        return None, None
    if '$' not in fullGlobalID:
        fullGlobalID = 'master$' + fullGlobalID
    keyAlias, _, customerID = fullGlobalID.partition('$')
    return keyAlias, customerID, dataID


def IsCanonicalVersion(versionName):
    """
    Check given ``versionName`` to have a valid format.
    """
    return re.match('^F\d+?(AM|PM)\d*?$', versionName) is not None


def IsPacketNameCorrect(fileName):
    """
    Check the ``fileName`` (this is a last 3 parts of packet ID) to have a
    valid format.
    """
    return re.match('^\d+?\-\d+?\-(Data|Parity)$', fileName) is not None


def IsPathIDCorrect(pathID, customer_id_mandatory=False):
    """
    Validate a given ``pathID``, should have only digits and '/' symbol.
    """
    try:
        customerGlobalID, remotePath = SplitPacketID(pathID)
    except:
        return False
    if customer_id_mandatory:
        try:
            user, _, host = customerGlobalID.rpartition('@')
        except:
            return False
        if not user:
            return False
        if not host:
            return False
    # TODO: more strict validation
    return remotePath.replace('/', '').isdigit()


def IsBackupIDCorrect(backupID, customer_id_mandatory=False):
    """
    Validate a given ``backupID``, must have such format:

        alice@idhost.org:0/0/1/0/F20131120053803PM.
    """
    pathID, _, version = backupID.rpartition('/')
    if not IsCanonicalVersion(version):
        return False
    if not IsPathIDCorrect(pathID, customer_id_mandatory=customer_id_mandatory):
        return False
    # TODO: more strict validation
    return True


def IsGlobalPathCorrect(globPath, customer_id_mandatory=False):
    """
    Validate a given ``globPath``, must have such format:

        some_key$alice@idhost.org:myfiles/flowers/cactus.png
    """
    customerGlobalID, remotePath = SplitPacketID(globPath)
    if customer_id_mandatory:
        from bitdust.userid import global_id
        return global_id.IsValidGlobalUser(customerGlobalID)
    if not remotePath:
        return False
    parts = remotePath.split('/')
    if len(parts) > 50:
        return False
    # TODO: more strict validation
    return True


def BidBnSnDp(packetID):
    """
    A wrapper for ``Split()`` method, returns tuple of 4 items.
    """
    tupl5 = Split(packetID)
    return MakeBackupID(tupl5[0], tupl5[1]), tupl5[2], tupl5[3], tupl5[4]


def UsrBidBnSnDp(packetID):
    """
    Another wrapper for ``Split()`` method.
    """
    return Split(packetID)


def KeyAlias(inp, normalize_key_alias=True):
    if not inp:
        return None
    customerGlobalID = inp
    if ':' in inp:
        try:
            customerGlobalID, _, _ = inp.partition(':')
        except:
            return None
    if '$' not in customerGlobalID and normalize_key_alias:
        customerGlobalID = 'master$' + customerGlobalID
    keyAlias, _, _ = customerGlobalID.partition('$')
    return str(keyAlias)


def CustomerIDURL(backupID, as_field=True):
    """
    A wrapper for ``Split()`` method to get customer idurl from backup ID.
    """
    user, _, _ = backupID.strip().rpartition(':')
    if not user:
        from bitdust.userid import my_id
        if as_field:
            return my_id.getIDURL()
        return my_id.getIDURL().to_bin()
    from bitdust.userid import global_id
    return global_id.GlobalUserToIDURL(user, as_field=as_field)


def KeyAliasCustomer(backupID, as_field=True):
    """
    A wrapper for ``Split()`` method to get key alias and customer idurl from backup ID.
    """
    user, _, _ = backupID.strip().rpartition(':')
    if not user:
        from bitdust.userid import my_id
        if as_field:
            return 'master', my_id.getIDURL()
        return 'master', my_id.getIDURL().to_bin()
    key_alias, _, user = user.strip().partition('$')
    from bitdust.userid import global_id
    return key_alias, global_id.GlobalUserToIDURL(user, as_field=as_field)


def RemotePath(backupID):
    """
    A wrapper for ``Split()`` method to get only remote_path from backup ID.
    """
    _, _, remote_path = backupID.strip().rpartition(':')
    if not remote_path:
        return ''
    return remote_path


def BackupID(packetID):
    """
    A wrapper for ``Split()`` method to get the first part - backup ID.
    """
    customerGlobalID, remotePathWithVersion, _, _, _ = Split(packetID)[1]
    return MakeBackupID(customer=customerGlobalID, path_id=remotePathWithVersion)


def BlockNumber(packetID):
    """
    A wrapper for ``Split()`` method to get the second part - block number.
    """
    return Split(packetID)[2]


def SupplierNumber(packetID):
    """
    A wrapper for ``Split()`` method to get the third part - supplier number.
    """
    return Split(packetID)[3]


def DataOrParity(packetID):
    """
    A wrapper for ``Split()`` method to get the last part - is this Data or Parity packet .
    """
    return Split(packetID)[4]


def parentPathsList(ID):
    """
    Return an iterator to go thru all parent paths of the given path ``ID``:
    list(packetid.parentPathsList('0/0/1/0/F20131120053803PM/0-1-Data'))

    ['0', '0/0', '0/0/1', '0/0/1/0', '0/0/1/0/F20131120053803PM', '0/0/1/0/F20131120053803PM/0-1-Data']
    """
    path = ''
    for word in ID.split('/'):
        if path:
            path += '/'
        path += word
        yield path


def LatestBackupID(backupID):
    """
    Create IDURL object from input key_id and return new key_id (with same key_alias) from that IDURL object.
    This way you can be sure that given key_id is pointing to the correct owner IDURL.
    """
    if not backupID:
        return backupID
    from bitdust.userid import global_id
    glob_id = global_id.ParseGlobalID(backupID, as_field=True)
    if not glob_id['idurl']:
        from bitdust.logs import lg
        lg.err('invalid backupID: %r' % backupID)
        return backupID
    return global_id.MakeGlobalID(
        key_id=glob_id['key_id'],
        path=glob_id['path'],
        version=glob_id['version'],
    )


def MakeQueueMessagePacketID(queue_id, unique_id):
    """
    Prepare text string to be used as PacketID when sending queue message.
    """
    return 'queue_{}_{}'.format(queue_id, unique_id)


def SplitQueueMessagePacketID(packet_id):
    """
    Extract `queue_id` and `unique_id` from PacketID expecting to have queue_message format.
    """
    if packet_id.startswith('queue_'):
        packet_id = packet_id[6:]
    queue_id, _, unique_id = packet_id.rpartition('_')
    return queue_id, unique_id



import threading
import asyncio


class HyperlaneManager(threading.Thread):
    def __init__(self, app_log):
        threading.Thread.__init__(self, target=self.loop_in_thread, args=(self.loop, ))
        self.app_log = app_log

    async def hyperlane_manager(self):
        self.app_log.warning('Hyperlane manager initiated')
        while True:
            self.app_log.warning('Hyperlane manager running')
            await asyncio.sleep(5)

    def loop_in_thread(self, loop):
        asyncio.set_event_loop(loop)
        loop.run_until_complete(self.hyperlane_manager())

    loop = asyncio.get_event_loop()


if __name__ == '__main__':
    import options
    import log

    config = options.Get()
    config.read()
    app_log = log.log('hyperlane.log', 'WARNING', True)

    #wm = HyperlaneManager(app_log)
    #hyperlane_thread = threading.Thread(target=wm.loop_in_thread, args=(wm.loop,))
    #hyperlane_manager = HyperlaneManager(app_log)
    #hyperlane_manager.start()
    hyperlane_manager = HyperlaneManager(app_log)
    hyperlane_manager.start()

    print('we can continue without being blocked')



#!/usr/bin/python
# service_data_motion.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_data_motion.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_data_motion
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return DataMotionService()


class DataMotionService(LocalService):

    service_name = 'service_data_motion'
    config_path = 'services/data-motion/enabled'

    def dependent_on(self):
        return [
            'service_employer',
        ]

    def init(self, **kwargs):
        from bitdust.main import events
        events.add_subscriber(self._on_my_suppliers_all_hired, 'my-suppliers-all-hired')
        events.add_subscriber(self._on_my_suppliers_yet_not_hired, 'my-suppliers-yet-not-hired')

    def shutdown(self):
        from bitdust.main import events
        events.remove_subscriber(self._on_my_suppliers_yet_not_hired, 'my-suppliers-yet-not-hired')
        events.remove_subscriber(self._on_my_suppliers_all_hired, 'my-suppliers-all-hired')

    def start(self):
        from bitdust.logs import lg
        from bitdust.customer import fire_hire
        if not fire_hire.IsAllHired():
            lg.warn('service_data_motion() can not start right now, not all suppliers hired yet')
            return False
        from bitdust.main import events
        from bitdust.stream import io_throttle
        from bitdust.stream import data_sender
        from bitdust.stream import data_receiver
        io_throttle.init()
        data_sender.SetShutdownFlag(False)
        data_sender.A('init')
        data_receiver.A('init')
        events.add_subscriber(self._on_identity_url_changed, 'identity-url-changed')
        events.add_subscriber(self._on_supplier_modified, 'supplier-modified')
        return True

    def stop(self):
        from bitdust.main import events
        from bitdust.stream import io_throttle
        from bitdust.stream import data_sender
        from bitdust.stream import data_receiver
        events.remove_subscriber(self._on_supplier_modified, 'supplier-modified')
        events.remove_subscriber(self._on_identity_url_changed, 'identity-url-changed')
        data_receiver.A('shutdown')
        data_sender.SetShutdownFlag(True)
        data_sender.A('shutdown')
        io_throttle.shutdown()
        return True

    def health_check(self):
        return True

    def _on_my_suppliers_all_hired(self, evt):
        from bitdust.logs import lg
        from bitdust.services import driver
        if driver.is_enabled('service_data_motion'):
            if not driver.is_started('service_data_motion'):
                from bitdust.customer import fire_hire
                if fire_hire.IsAllHired():
                    lg.info('all my suppliers are hired, starting service_data_motion()')
                    driver.start_single('service_data_motion')

    def _on_my_suppliers_yet_not_hired(self, evt):
        from bitdust.logs import lg
        from bitdust.services import driver
        if driver.is_enabled('service_data_motion'):
            if driver.is_started('service_data_motion'):
                lg.info('my suppliers failed to hire, stopping service_data_motion()')
                driver.stop_single('service_data_motion')

    def _on_identity_url_changed(self, evt):
        from bitdust.logs import lg
        from bitdust.userid import id_url
        from bitdust.stream import io_throttle
        old_idurl = id_url.field(evt.data['old_idurl'])
        for supplier_idurl, supplier_queue in io_throttle.throttle().supplierQueues.items():
            if old_idurl == supplier_idurl:
                supplier_idurl.refresh()
                lg.info('found supplier idurl rotated in io_throttle: %r -> %r' % (evt.data['old_idurl'], evt.data['new_idurl']))
            if old_idurl == supplier_queue.customerIDURL:
                supplier_queue.customerIDURL.refresh()
                lg.info('found customer idurl rotated in io_throttle %r supplier queue: %r -> %r' % (supplier_idurl, evt.data['old_idurl'], evt.data['new_idurl']))

    def _on_supplier_modified(self, evt):
        from bitdust.stream import data_sender
        data_sender.A('restart')



#!/usr/bin/python
# id_url.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (id_url.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: id_url.

.. role:: red

This is a core module.

IDURL of your node can changed over time, see id_rotator() state machine which does that.
So you must remember your "old" identity sources and store locally old IDURL's you had.
This way when you receive a packet from another node which do not know your new IDURL yet,
you can understand that he is actually trying to contact you and act properly.

Also if you stored some packet for user A and then his IDURL changed you must still be able to
correctly reply to him and send stored data back when he requests.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

import os
import sys
import tempfile
import traceback

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio
from bitdust.system import local_fs

from bitdust.crypt import hashes

from bitdust.lib import strng
from bitdust.lib import nameurl

from bitdust.main import settings

#------------------------------------------------------------------------------

_IdentityHistoryDir = None
# TODO: if this dictionary grow too much use CodernityDB instead of in-memory storage
_KnownUsers = {}
_KnownIDURLs = {}
_MergedIDURLs = {}
_KnownSources = {}
_KnownUniqueNames = {}
_Ready = False

#------------------------------------------------------------------------------


def init():
    global _IdentityHistoryDir
    global _KnownUsers
    global _KnownIDURLs
    global _MergedIDURLs
    global _KnownSources
    global _KnownUniqueNames
    global _Ready
    from bitdust.userid import identity
    if _Debug:
        lg.out(_DebugLevel, 'id_url.init')
    if not _IdentityHistoryDir:
        _IdentityHistoryDir = settings.IdentityHistoryDir()
    if not os.path.exists(_IdentityHistoryDir):
        bpio._dir_make(_IdentityHistoryDir)
        lg.info('created new folder %r' % _IdentityHistoryDir)
    else:
        lg.info('using existing folder %r' % _IdentityHistoryDir)
    for_cleanup = []
    for one_user_dir in os.listdir(_IdentityHistoryDir):
        one_user_dir_path = os.path.join(_IdentityHistoryDir, one_user_dir)
        if not os.path.isdir(one_user_dir_path):
            continue
        one_user_identity_files = []
        for one_filename in os.listdir(one_user_dir_path):
            try:
                one_ident_number = int(one_filename)
            except:
                lg.exc()
                continue
            one_user_identity_files.append(one_ident_number)
        if _Debug:
            lg.out(_DebugLevel, 'id_url.init   found %d historical records in %r' % (len(one_user_identity_files), one_user_dir_path))
        one_user_identity_files.sort()
        for one_ident_file in one_user_identity_files:
            one_ident_path = os.path.join(one_user_dir_path, strng.to_text(one_ident_file))
            try:
                xmlsrc = local_fs.ReadTextFile(one_ident_path)
                known_id_obj = identity.identity(xmlsrc=xmlsrc)
                if not known_id_obj.isCorrect():
                    raise Exception('identity history in %r is broken, identity is not correct: %r' % (one_user_dir, one_ident_path))
                if not known_id_obj.Valid():
                    raise Exception('identity history in %r is broken, identity is not valid: %r' % (one_user_dir, one_ident_path))
            except Exception as exc:
                lg.warn(str(exc))
                for_cleanup.append(one_ident_path)
                continue
            one_pub_key = known_id_obj.getPublicKey()
            one_revision = known_id_obj.getRevisionValue()
            if one_pub_key not in _KnownUsers:
                _KnownUsers[one_pub_key] = one_user_dir_path
            one_unique_name = '{}_{}'.format(
                strng.to_text(one_ident_file).strip()[0:-4],
                strng.to_text(hashes.sha1(one_pub_key, hexdigest=True)),
            )
            known_sources = known_id_obj.getSources(as_originals=True)
            for known_idurl in reversed(known_sources):
                if known_idurl not in _KnownIDURLs:
                    _KnownIDURLs[known_idurl] = known_id_obj.getPublicKey()
                    if _Debug:
                        lg.out(_DebugLevel, '    new IDURL added: %r' % known_idurl)
                else:
                    if _KnownIDURLs[known_idurl] != known_id_obj.getPublicKey():
                        _KnownIDURLs[known_idurl] = known_id_obj.getPublicKey()
                        lg.warn('another user had same identity source: %r' % known_idurl)
                if one_pub_key not in _MergedIDURLs:
                    _MergedIDURLs[one_pub_key] = {}
                    if _Debug:
                        lg.out(_DebugLevel, '    new Public Key added: %s...' % one_pub_key[-10:])
                if one_revision in _MergedIDURLs[one_pub_key]:
                    if _MergedIDURLs[one_pub_key][one_revision] != known_idurl:
                        if _MergedIDURLs[one_pub_key][one_revision] not in known_sources:
                            lg.warn('rewriting existing identity revision %d : %r -> %r' % (one_revision, _MergedIDURLs[one_pub_key][one_revision], known_idurl))
                    _MergedIDURLs[one_pub_key][one_revision] = known_idurl
                else:
                    _MergedIDURLs[one_pub_key][one_revision] = known_idurl
                    if _Debug:
                        lg.out(_DebugLevel, '        revision %d merged with other %d known items' % (one_revision, len(_MergedIDURLs[one_pub_key])))
                if one_pub_key not in _KnownSources:
                    _KnownSources[one_pub_key] = []
                if one_unique_name not in _KnownUniqueNames:
                    _KnownUniqueNames[one_unique_name] = []
                for one_source in known_id_obj.getSources(as_originals=True):
                    if one_source not in _KnownSources[one_pub_key]:
                        _KnownSources[one_pub_key].append(one_source)
                        if _Debug:
                            lg.out(_DebugLevel, '    new source %r added for %r' % (one_source, one_pub_key[-10:]))
                    if one_source not in _KnownUniqueNames[one_unique_name]:
                        _KnownUniqueNames[one_unique_name].append(one_source)
                        if _Debug:
                            lg.out(_DebugLevel, '    new source %r added for unique name %r' % (one_source, one_unique_name))
    for one_ident_path in for_cleanup:
        if os.path.isfile(one_ident_path):
            lg.warn('about to erase broken historical identity file: %r' % one_ident_path)
            try:
                os.remove(one_ident_path)
            except:
                lg.exc()
    _Ready = True


def shutdown():
    global _IdentityHistoryDir
    global _KnownIDURLs
    global _KnownUsers
    global _Ready
    global _MergedIDURLs
    global _KnownSources
    global _KnownUniqueNames
    _IdentityHistoryDir = None
    _KnownUsers.clear()
    _KnownIDURLs.clear()
    _MergedIDURLs.clear()
    _KnownSources.clear()
    _KnownUniqueNames.clear()
    _Ready = False


#------------------------------------------------------------------------------


def known():
    global _KnownIDURLs
    return _KnownIDURLs


def merged(pub_key=None):
    global _MergedIDURLs
    if pub_key is None:
        return _MergedIDURLs
    return _MergedIDURLs.get(pub_key, {})


def users(pub_key=None):
    global _KnownUsers
    if pub_key is None:
        return _KnownUsers
    return _KnownUsers.get(pub_key, None)


def sources(pub_key=None):
    global _KnownSources
    if pub_key is None:
        return _KnownSources
    return _KnownSources.get(pub_key, [])


def unique_names(unique_name=None):
    global _KnownUniqueNames
    if unique_name is None:
        return _KnownUniqueNames
    return _KnownUniqueNames.get(unique_name, [])


#------------------------------------------------------------------------------


def identity_cached(new_id_obj):
    """
    After receiving identity file of another user we need to check his identity sources.
    I can be file from identity server or Identity() packet received directly from remote peer.
    Also it can be my own identity that was changed locally.
    In any case we need to take certain actions if those identity sources changed.
    First identity source forms IDURL of that identity and act as unique global ID of that BitDust node.
    When first identity source changed (because identity server went down) identity is "rotated":
    second identity source will be placed on the first position and IDURL will change.
    In that case we need to remember new IDURL and keep track of old IDURL of that user - this way we can
    match and merge different IDURL's for one owner.
    """
    global _IdentityHistoryDir
    global _KnownUsers
    global _KnownIDURLs
    global _MergedIDURLs
    global _KnownSources
    global _KnownUniqueNames
    from bitdust.userid import identity
    pub_key = new_id_obj.getPublicKey()
    user_name = new_id_obj.getIDName()
    if _Debug:
        lg.args(_DebugLevel, user_name=user_name)
    is_identity_rotated = False
    latest_id_obj = None
    latest_sources = []
    for_cleanup = []
    if pub_key not in _KnownUsers:
        user_path = tempfile.mkdtemp(prefix=user_name + '@', dir=_IdentityHistoryDir)
        _KnownUsers[pub_key] = user_path
        first_identity_file_path = os.path.join(user_path, '0')
        if os.path.exists(first_identity_file_path):
            try:
                os.remove(first_identity_file_path)
            except:
                lg.exc()
        local_fs.WriteBinaryFile(first_identity_file_path, new_id_obj.serialize())
        if _Debug:
            lg.out(_DebugLevel, 'id_url.identity_cached wrote first item for user %r in identity history: %r' % (user_name, first_identity_file_path))
    else:
        user_path = _KnownUsers[pub_key]
        user_identity_files = sorted(map(int, os.listdir(user_path)))
        if len(user_identity_files) == 0:
            lg.warn('identity history for user %r is broken, public key is known, but no identity files found' % user_name)
        latest_identity_file_path = ''
        latest_pub_key = None
        latest_revision = -1
        known_revisions = set()
        for id_file in user_identity_files:
            identity_file_path = os.path.join(user_path, strng.to_text(id_file))
            xmlsrc = local_fs.ReadBinaryFile(identity_file_path)
            one_id_obj = identity.identity(xmlsrc=xmlsrc)
            if not one_id_obj.isCorrect():
                lg.warn('identity history for user %r is broken, identity in the file %r is not correct' % (user_name, identity_file_path))
                for_cleanup.append(identity_file_path)
                continue
            if not one_id_obj.Valid():
                lg.warn('identity history for user %r is broken, identity in the file %r is not valid' % (user_name, identity_file_path))
                for_cleanup.append(identity_file_path)
                continue
            if not latest_pub_key:
                latest_pub_key = one_id_obj.getPublicKey()
            if latest_pub_key != one_id_obj.getPublicKey():
                lg.err('identity history for user %r is broken, public key not matching in the file %r' % (user_name, identity_file_path))
                for_cleanup.append(identity_file_path)
                continue
            known_revisions.add(one_id_obj.getRevisionValue())
            if one_id_obj.getRevisionValue() > latest_revision:
                latest_revision = one_id_obj.getRevisionValue()
                latest_identity_file_path = identity_file_path
        xmlsrc = local_fs.ReadBinaryFile(latest_identity_file_path)
        if xmlsrc:
            latest_id_obj = identity.identity(xmlsrc=xmlsrc)
            if latest_id_obj.getPublicKey() != new_id_obj.getPublicKey():
                raise Exception('identity history for user %r is broken, public key not matching' % user_name)
            if latest_id_obj.getIDName() != new_id_obj.getIDName():
                lg.warn('found another user name in identity history for user %r : %r' % (user_name, latest_id_obj.getIDName()))
            if new_id_obj.getRevisionValue() in known_revisions:
                if _Debug:
                    lg.out(_DebugLevel, 'id_url.identity_cached revision %d already known for user %r' % (new_id_obj.getRevisionValue(), user_name))
            else:
                latest_sources = latest_id_obj.getSources(as_originals=True)
                new_sources = new_id_obj.getSources(as_originals=True)
                if latest_sources == new_sources:
                    if os.path.exists(latest_identity_file_path):
                        try:
                            os.remove(latest_identity_file_path)
                        except:
                            lg.exc()
                    local_fs.WriteBinaryFile(latest_identity_file_path, new_id_obj.serialize())
                    if _Debug:
                        lg.out(_DebugLevel, 'id_url.identity_cached latest identity sources for user %r did not changed, updated file %r' % (user_name, latest_identity_file_path))
                else:
                    next_identity_file = user_identity_files[-1] + 1
                    next_identity_file_path = os.path.join(user_path, strng.to_text(next_identity_file))
                    if os.path.exists(next_identity_file_path):
                        try:
                            os.remove(next_identity_file_path)
                        except:
                            lg.exc()
                    local_fs.WriteBinaryFile(next_identity_file_path, new_id_obj.serialize())
                    is_identity_rotated = True
                    if _Debug:
                        lg.out(_DebugLevel, 'id_url.identity_cached identity sources for user %r changed, wrote new item in the history: %r' % (user_name, next_identity_file_path))
    new_revision = new_id_obj.getRevisionValue()
    new_sources = new_id_obj.getSources(as_originals=True)
    for new_idurl in reversed(new_sources):
        if new_idurl not in _KnownIDURLs:
            _KnownIDURLs[new_idurl] = new_id_obj.getPublicKey()
            if _Debug:
                lg.out(_DebugLevel, 'id_url.identity_cached new IDURL added: %r' % new_idurl)
        else:
            if _KnownIDURLs[new_idurl] != new_id_obj.getPublicKey():
                lg.warn('another user had same identity source: %r' % new_idurl)
                _KnownIDURLs[new_idurl] = new_id_obj.getPublicKey()
        if pub_key not in _MergedIDURLs:
            _MergedIDURLs[pub_key] = {}
            if _Debug:
                lg.out(_DebugLevel, 'id_url.identity_cached new Public Key added: %s...' % pub_key[-10:])
        prev_idurl = _MergedIDURLs[pub_key].get(new_revision, None)
        if new_revision in _MergedIDURLs[pub_key]:
            if _MergedIDURLs[pub_key][new_revision] != new_idurl:
                if nameurl.GetName(_MergedIDURLs[pub_key][new_revision]) == nameurl.GetName(new_idurl):
                    if _MergedIDURLs[pub_key][new_revision] not in new_sources:
                        lg.warn('rewriting existing identity revision %d : %r -> %r' % (new_revision, _MergedIDURLs[pub_key][new_revision], new_idurl))
            _MergedIDURLs[pub_key][new_revision] = new_idurl
        else:
            _MergedIDURLs[pub_key][new_revision] = new_idurl
            if _Debug:
                lg.out(_DebugLevel, 'id_url.identity_cached added new revision %d for user %r, total revisions %d: %r -> %r' % (new_revision, user_name, len(_MergedIDURLs[pub_key]), prev_idurl, new_idurl))
    if pub_key not in _KnownSources:
        _KnownSources[pub_key] = []
    for one_source in new_sources:
        if one_source not in _KnownSources[pub_key]:
            _KnownSources[pub_key].append(one_source)
            if _Debug:
                lg.out(_DebugLevel, 'id_url.identity_cached added new source %r for user %r' % (one_source, user_name))
    unique_name = '{}_{}'.format(
        user_name,
        strng.to_text(hashes.sha1(pub_key, hexdigest=True)),
    )
    if unique_name not in _KnownUniqueNames:
        _KnownUniqueNames[unique_name] = []
    for one_source in new_sources:
        if one_source not in _KnownUniqueNames[unique_name]:
            _KnownUniqueNames[unique_name].append(one_source)
            if _Debug:
                lg.out(_DebugLevel, 'id_url.identity_cached added new source %r for unique name %r' % (one_source, unique_name))
    if _Debug:
        lg.args(_DebugLevel, is_identity_rotated=is_identity_rotated, latest_id_obj=bool(latest_id_obj))
    if is_identity_rotated and latest_id_obj is not None:
        latest_revision = latest_id_obj.getRevisionValue()
        if _Debug:
            lg.args(_DebugLevel, new_revision=new_revision, latest_revision=latest_revision)
        if new_revision > latest_revision:
            lg.info('found rotated identity after caching %r -> %r' % (latest_id_obj.getSources(as_originals=True)[0], new_sources[0]))
            from bitdust.main import events
            events.send('identity-rotated', data=dict(
                old_idurls=latest_id_obj.getSources(as_originals=True),
                new_idurls=new_id_obj.getSources(as_originals=True),
                old_revision=latest_id_obj.getRevisionValue(),
                new_revision=new_revision,
            ))
            if latest_id_obj.getIDURL(as_original=True) != new_id_obj.getIDURL(as_original=True):
                events.send('identity-url-changed', data=dict(
                    old_idurl=latest_id_obj.getIDURL(as_original=True),
                    new_idurl=new_id_obj.getIDURL(as_original=True),
                    old_revision=latest_id_obj.getRevisionValue(),
                    new_revision=new_revision,
                ))
            from bitdust.userid import my_id
            if my_id.isLocalIdentityReady():
                if my_id.getIDURL() == new_id_obj.getIDURL():
                    # yapf: disable
                    events.send('my-identity-rotated', data=dict(
                        old_idurls=latest_id_obj.getSources(as_originals=True),
                        new_idurls=new_id_obj.getSources(as_originals=True),
                        old_revision=latest_id_obj.getRevisionValue(),
                        new_revision=new_revision,
                    ))
                    # yapf: enable
                    if latest_id_obj.getIDURL(as_original=True) != new_id_obj.getIDURL(as_original=True):
                        # yapf: disable
                        events.send('my-identity-url-changed', data=dict(
                            old_idurl=latest_id_obj.getIDURL(as_original=True),
                            new_idurl=new_id_obj.getIDURL(as_original=True),
                            old_revision=latest_id_obj.getRevisionValue(),
                            new_revision=new_revision,
                        ))
                        # yapf: enable
        else:
            lg.warn('cached out-dated revision %d for %r, most recent revision is %d' % (new_revision, new_sources[0], latest_revision))
    else:
        if _Debug:
            lg.out(_DebugLevel, 'id_url.identity_cached revision %d for %r' % (new_revision, new_sources[0]))
    for identity_file_path in for_cleanup:
        if os.path.isfile(identity_file_path):
            try:
                os.remove(identity_file_path)
            except:
                lg.exc()
    return True


#------------------------------------------------------------------------------


def field(idurl):
    """
    Translates string into `ID_URL_FIELD` object.
    Also we try to read from local identity cache folder if do not know given "idurl".
    """
    global _KnownIDURLs
    if isinstance(idurl, ID_URL_FIELD):
        return idurl
    if idurl in [None, 'None', '', b'None', b'', False]:
        return ID_URL_FIELD(idurl)
    idurl = strng.to_bin(idurl.strip())
    if idurl not in _KnownIDURLs:
        if _Debug:
            lg.out(_DebugLevel, 'id_url.field   will try to find %r in local identity cache' % idurl)
        from bitdust.contacts import identitydb
        cached_ident = identitydb.get_ident(idurl)
        if cached_ident:
            identity_cached(cached_ident)
        else:
            if _Debug:
                cod = sys._getframe(1).f_back.f_code
                modul = os.path.basename(cod.co_filename).replace('.py', '')
                caller = cod.co_name
                lg.warn('unknown yet idurl %s, call from %s.%s' % (idurl, modul, caller))
    return ID_URL_FIELD(idurl)


def fields_list(idurl_list):
    """
    Same as `field()` but for lists.
    """
    return list(map(field, idurl_list))


def fields_dict(idurl_dict):
    """
    Translates dictionary keys into `ID_URL_FIELD` objects.
    """
    return {field(k): v for k, v in idurl_dict.items()}


def is_idurl(value):
    """
    Return True if input is `ID_URL_FIELD` field or string in valid format.
    """
    if value in [None, 'None', '', b'None', b'', False]:
        return False
    if isinstance(value, ID_URL_FIELD):
        return True
    if not strng.is_string(value):
        return False
    v = strng.to_text(value)
    if not v.startswith('http') or not v.endswith('.xml') or v.count('://') != 1:
        return False
    return True


def to_bin(idurl):
    """
    Translates `ID_URL_FIELD` to binary string.
    """
    if isinstance(idurl, ID_URL_FIELD):
        return idurl.to_bin()
    if idurl in [None, 'None', '', b'None', b'', False]:
        return b''
    return strng.to_bin(idurl)


def to_original(idurl):
    """
    Translates `ID_URL_FIELD` to binary string, but returns its original value.
    """
    if isinstance(idurl, ID_URL_FIELD):
        return idurl.original()
    if idurl in [None, 'None', '', b'None', b'', False]:
        return b''
    return strng.to_bin(idurl)


def to_list(iterable_object, as_field=True, as_bin=False, as_original=False):
    """
    Creates list of desired objects from given `iterable_object`.
    """
    if not iterable_object:
        return iterable_object
    if as_original:
        return list(map(to_original, iterable_object))
    if as_field:
        return list(map(field, iterable_object))
    if as_bin:
        return list(map(to_bin, iterable_object))
    return list(map(strng.to_text, iterable_object))


def to_bin_list(iterable_object):
    """
    Just an alias for to_list().
    """
    if not iterable_object:
        return iterable_object
    return to_list(iterable_object, as_field=False, as_bin=True)


def to_original_list(iterable_object):
    """
    Just an alias for to_list() to extract original values from idurl's.
    """
    if not iterable_object:
        return iterable_object
    return to_list(iterable_object, as_original=True)


def to_bin_dict(idurl_dict):
    """
    Translates dictionary keys into binary strings if they were `ID_URL_FIELD` objects.
    """
    return {to_bin(k): v for k, v in idurl_dict.items()}


def is_in(idurl, iterable_object, as_field=True, as_bin=False):
    """
    Equivalent of `idurl in iterable_object`.
    Because it can be that you need to compare not `ID_URL_FIELD` objects in memory but normal strings.
    So then you will prefer to avoid translating into field object.
    """
    if not iterable_object:
        return False
    if as_field:
        return field(idurl) in fields_list(iterable_object)
    if as_bin:
        return to_bin(idurl) in to_bin_list(iterable_object)
    return strng.to_text(idurl) in to_list(iterable_object, as_field=False, as_bin=False)


def is_not_in(idurl, iterable_object, as_field=True, as_bin=False):
    """
    Vise-versa of `is_in()` method.
    """
    return not is_in(idurl=idurl, iterable_object=iterable_object, as_field=as_field, as_bin=as_bin)


def get_from_dict(idurl, dict_object, default=None, as_field=True, as_bin=False):
    """
    Equivalent of `{<some dict>}.get(idurl, default)`.
    Because it can be that you need to keep not `ID_URL_FIELD` objects in your dictionary but a normal strings.
    So then you will prefer to avoid translating into field object.
    """
    if as_field:
        return dict_object.get(field(idurl), default)
    if as_bin:
        return dict_object.get(to_bin(idurl), default)
    return dict_object.get(strng.to_text(idurl), default)


def is_cached(idurl):
    """
    Return True if given identity was already cached - that means I know his public key.
    """
    global _KnownIDURLs
    if isinstance(idurl, ID_URL_FIELD):
        idurl = idurl.to_bin()
    else:
        idurl = strng.to_bin(idurl)
    cached = idurl in _KnownIDURLs
    if not cached:
        if _Debug:
            lg.args(_DebugLevel, idurl=idurl, cached=cached)
    return cached


def is_empty(idurl):
    """
    Return True if given input is None, empty string or empty ID_URL_FIELD object.
    """
    if isinstance(idurl, ID_URL_FIELD):
        return not bool(idurl)
    if idurl in [None, 'None', '', b'None', b'', False]:
        return True
    return not bool(idurl)


def is_some_empty(iterable_object):
    """
    Returns True if given iterable_object contains some empty idurl field.
    """
    return is_in(ID_URL_FIELD(b''), iterable_object=iterable_object, as_field=True, as_bin=False)


def is_the_same(idurl1, idurl2):
    """
    Compares two ID_URL_FIELD objects or binary strings taking in account if those IDURLs are already cached.
    """
    idurl1 = field(idurl1)
    idurl2 = field(idurl2)
    if not idurl1 or not idurl2:
        return False
    if is_cached(idurl1) and is_cached(idurl2):
        return idurl1 == idurl2
    return idurl1.to_bin() == idurl2.to_bin()


def empty_count(iterable_object):
    """
    Returns number of empty idurl fields or empty strings found in given `iterable_object`.
    """
    count = 0
    for idurl in iterable_object:
        if is_empty(idurl):
            count += 1
    return count


def get_latest_revision(idurl):
    """
    Return latest known IDURL (as binary string) and revision number for given idurl field or string.
    This info is extracted from in-memory "index" of all known identity objects.
    """
    global _MergedIDURLs
    global _KnownIDURLs
    latest_rev = -1
    latest_idurl = b''
    idurl_bin = to_bin(idurl)
    if not idurl_bin:
        return latest_idurl, latest_rev
    if idurl_bin not in _KnownIDURLs:
        return latest_idurl, latest_rev
    pub_key = _KnownIDURLs[idurl_bin]
    if pub_key not in _MergedIDURLs:
        lg.warn('idurl %r is known but has no known revisions yet' % idurl_bin)
        return latest_idurl, latest_rev
    for rev, another_idurl in _MergedIDURLs[pub_key].items():
        if rev >= latest_rev:
            latest_idurl = another_idurl
            latest_rev = rev
    return latest_idurl, latest_rev


def get_latest_ident(pub_key):
    global _KnownUsers
    from bitdust.userid import identity
    user_path = _KnownUsers.get(pub_key)
    if not user_path:
        return None
    user_identity_files = sorted(map(int, os.listdir(user_path)))
    if len(user_identity_files) == 0:
        lg.warn('identity history is broken, public key is known, but no identity files found')
    latest_revision = -1
    latest_ident = None
    known_revisions = set()
    for_cleanup = []
    for id_file in user_identity_files:
        identity_file_path = os.path.join(user_path, strng.to_text(id_file))
        xmlsrc = local_fs.ReadBinaryFile(identity_file_path)
        one_id_obj = identity.identity(xmlsrc=xmlsrc)
        if not one_id_obj.isCorrect():
            lg.warn('identity history is broken, identity in the file %r is not correct' % identity_file_path)
            for_cleanup.append(identity_file_path)
            continue
        if not one_id_obj.Valid():
            lg.warn('identity history is broken, identity in the file %r is not valid' % identity_file_path)
            for_cleanup.append(identity_file_path)
            continue
        if pub_key != one_id_obj.getPublicKey():
            lg.err('identity history is broken, public key not matching in the file %r' % identity_file_path)
            for_cleanup.append(identity_file_path)
            continue
        known_revisions.add(one_id_obj.getRevisionValue())
        if one_id_obj.getRevisionValue() > latest_revision:
            latest_revision = one_id_obj.getRevisionValue()
            latest_ident = one_id_obj
    return latest_ident


def list_known_idurls(idurl, num_revisions=5, include_revisions=False):
    """
    Return latest known revisions of given `idurl` (as binary strings) and revision numbers as a list of tuples.
    This info is extracted from in-memory "index" of all known identity objects.
    """
    global _MergedIDURLs
    global _KnownIDURLs
    idurl_bin = to_bin(idurl)
    if not idurl_bin:
        if _Debug:
            lg.args(_DebugLevel, idurl=idurl, ret=idurl_bin)
        if include_revisions:
            return [(idurl_bin, -1)]
        return [idurl_bin]
    if idurl_bin not in _KnownIDURLs:
        if _Debug:
            lg.args(_DebugLevel, idurl=idurl, ret=idurl_bin)
        if include_revisions:
            return [(idurl_bin, -1)]
        return [idurl_bin]
    pub_key = _KnownIDURLs[idurl_bin]
    if pub_key not in _MergedIDURLs:
        lg.warn('idurl %r does not have any known revisions' % idurl_bin)
        if include_revisions:
            return [(idurl_bin, -1)]
        return [idurl_bin]
    known_revisions = sorted(_MergedIDURLs[pub_key].keys(), reverse=True)
    results = []
    for rev in known_revisions:
        another_idurl = _MergedIDURLs[pub_key][rev]
        if include_revisions:
            if another_idurl not in [i[0] for i in results]:
                results.append((another_idurl, rev))
        else:
            if another_idurl not in results:
                results.append(another_idurl)
        if len(results) >= num_revisions:
            break
    if _Debug:
        lg.args(_DebugLevel, idurl=idurl, ret=results)
    return results


def idurl_to_id(idurl_text, by_parts=False):
    """
    Translates raw IDURL string (not ID_URL_FIELD) into global id short form:
        "http://somehost.com/alice.xml" -> "alice@somehost.com"
    """
    if not idurl_text:
        if by_parts:
            return '', ''
        return idurl_text
    _, host, port, _, filename = nameurl.UrlParseFast(idurl_text)
    if filename.count('.'):
        username = filename.split('.')[0]
    else:
        username = filename
    if port:
        host = '%s_%s' % (host, port)
    if by_parts:
        return username, host
    return '%s@%s' % (username, host)


#------------------------------------------------------------------------------


class ID_URL_FIELD(object):

    """
    A class represents a valid, verified and synced IDURL identifier of a device.
    The IDURL is always corresponding to the identity file.

    When your identity file is updated due to rotation to another identity server,
    the "host" component of your IDURL is changed:
        `http://old-server/alice.xml` will become `http://new-server.org/alice.xml`

    The IDURL can also be presented in a shorter form, which is also called global user ID, so:
        `alice@old-server.com` will become `alice@new-server.org`

    All that is handled automatically and your instance of `ID_URL_FIELD` will receive updated info automatically.
    When incoming identity files are first received, processed and cached your `ID_URL_FIELD` variable is
    automatically synchornized using `ID_URL_FIELD.refresh()` method.

    This is why you can always trust and be able to compare two user IDs and verify recipient/sender identity.
    """

    def __init__(self, idurl):
        self.current = b''
        self.current_as_string = ''
        self.current_id = ''
        self.latest = b''
        self.latest_as_string = ''
        self.latest_id = ''
        self.latest_revision = -1
        self._unique_name = ''
        if isinstance(idurl, ID_URL_FIELD):
            self.current = idurl.current
        else:
            if idurl in [None, 'None', '', b'None', b'', False]:
                self.current = b''
            else:
                self.current = strng.to_bin(idurl.strip())
        self.current_as_string = strng.to_text(self.current)
        username, current_host = idurl_to_id(self.current, by_parts=True)
        self.username = username
        self.current_host = current_host
        self.current_id = '%s@%s' % (self.username, self.current_host)
        self.latest, self.latest_revision = get_latest_revision(self.current)
        if not self.latest:
            self.latest = self.current
            self.latest_revision = -1
        self.latest_as_string = strng.to_text(self.latest)
        latest_username, latest_host = idurl_to_id(self.latest, by_parts=True)
        if latest_username != self.username:
            caller_code = sys._getframe().f_back.f_code
            caller_method = caller_code.co_name
            caller_modul = os.path.basename(caller_code.co_filename).replace('.py', '')
            if caller_method.count('lambda') or caller_method == 'field':
                caller_method = sys._getframe(1).f_back.f_code.co_name
            exc = ValueError('tried to modify username of the identity %r -> %r' % (self.current, self.latest))
            lg.exc(msg='called from %s.%s()' % (caller_modul, caller_method), exc_value=exc)
            raise exc
        self.latest_host = latest_host
        self.latest_id = '%s@%s' % (self.username, self.latest_host)
        if _Debug:
            lg.out(_DebugLevel*2, 'NEW ID_URL_FIELD(%r) with id=%r latest=%r' % (self.current, id(self), self.latest))

    def __del__(self):
        try:
            if _Debug:
                lg.out(_DebugLevel*2, 'DELETED ID_URL_FIELD(%r) with id=%r latest=%r' % (self.current, id(self), self.latest))
        except:
            lg.exc()

    def __eq__(self, idurl):
        # always check type : must be `ID_URL_FIELD`
        if not isinstance(idurl, ID_URL_FIELD):
            # to be able to compare with empty value lets make an exception
            if idurl in [None, 'None', '', b'None', b'', False]:
                return not bool(self.latest)
            # in other cases must raise an exception
            caller_code = sys._getframe().f_back.f_code
            caller_method = caller_code.co_name
            caller_modul = os.path.basename(caller_code.co_filename).replace('.py', '')
            if caller_method.count('lambda'):
                caller_method = sys._getframe(1).f_back.f_code.co_name
            exc = TypeError('tried to compare ID_URL_FIELD(%r) with %r of type %r' % (self.latest_as_string, idurl, type(idurl)))
            lg.exc(msg='called from %s.%s()' % (caller_modul, caller_method), exc_value=exc)
            raise exc

        # could be empty field also
        if not idurl:
            return not bool(self.latest)

        # check if we know both sources
        my_pub_key = self.to_public_key(raise_error=False)
        other_pub_key = idurl.to_public_key(raise_error=False)
        if my_pub_key is None or other_pub_key is None:
            # if we do not know some of the sources - so can't be sure
            caller_code = sys._getframe().f_back.f_code
            caller_method = caller_code.co_name
            caller_modul = os.path.basename(caller_code.co_filename).replace('.py', '')
            if caller_method.count('lambda') or caller_method.startswith('_'):
                caller_method = sys._getframe(1).f_back.f_code.co_name
            if my_pub_key is None:
                exc = KeyError('unknown idurl %r in %s.%s' % (self.current, caller_modul, caller_method))
            else:
                exc = KeyError('unknown idurl %r in %s.%s' % (idurl.current, caller_modul, caller_method))
            lg.exc(msg='called from %s.%s()' % (caller_modul, caller_method), exc_value=exc)
            if _Debug:
                lg.out(_DebugLevel, traceback.format_list(traceback.format_stack()))
            raise exc

        # now compare based on public key
        result = (other_pub_key == my_pub_key)
        if _Debug:
            lg.args(_DebugLevel*2, idurl=idurl, current=self.current, latest=self.latest, result=result)
        return result

    def __ne__(self, idurl):
        # always check type : must be `ID_URL_FIELD`
        if not isinstance(idurl, ID_URL_FIELD):
            # to be able to compare with empty value lets make an exception
            if idurl in [None, 'None', '', b'None', b'', False]:
                return bool(self.latest)
            # in other cases must raise an exception
            caller_code = sys._getframe().f_back.f_code
            caller_method = caller_code.co_name
            caller_modul = os.path.basename(caller_code.co_filename).replace('.py', '')
            if caller_method.count('lambda') or caller_method.startswith('_'):
                caller_method = sys._getframe(1).f_back.f_code.co_name
            exc = TypeError('tried to compare ID_URL_FIELD(%r) with %r of type %r' % (self.latest_as_string, idurl, type(idurl)))
            lg.exc(msg='called from %s.%s()' % (caller_modul, caller_method), exc_value=exc)
            raise exc

        # could be empty field also
        if not idurl:
            return bool(self.latest)

        # check if we know both sources
        my_pub_key = self.to_public_key(raise_error=True)
        other_pub_key = idurl.to_public_key(raise_error=True)
        if my_pub_key is None or other_pub_key is None:
            # if we do not know some of the sources - so can't be sure
            caller_code = sys._getframe().f_back.f_code
            caller_method = caller_code.co_name
            caller_modul = os.path.basename(caller_code.co_filename).replace('.py', '')
            if caller_method.count('lambda') or caller_method.startswith('_'):
                caller_method = sys._getframe(1).f_back.f_code.co_name
            if my_pub_key is None:
                exc = KeyError('unknown idurl %r in %s.%s' % (self.current, caller_modul, caller_method))
            else:
                exc = KeyError('unknown idurl %r in %s.%s' % (idurl.current, caller_modul, caller_method))
            lg.exc(msg='called from %s.%s()' % (caller_modul, caller_method), exc_value=exc)
            if _Debug:
                lg.out(_DebugLevel, traceback.format_list(traceback.format_stack()))
            raise exc

        # now compare based on public key
        result = (other_pub_key != my_pub_key)
        if _Debug:
            lg.args(_DebugLevel*2, idurl=idurl, current=self.current, latest=self.latest, result=result)
        return result

    def __hash__(self):
        # this trick should make `idurl in some_dictionary` work correctly
        # if idurl1 and idurl2 are different sources of same identity they both must be matching
        # so it must never happen like that: (idurl1 in some_dictionary) and (idurl2 in some_dictionary)
        # same check you can do in a different way: `id_url.is_in(idurl, some_dictionary)`
        pub_key = self.to_public_key()
        hsh = pub_key.__hash__()
        if _Debug:
            lg.args(_DebugLevel*2, current=self.current, latest=self.latest, hash=hsh)
        return hsh

    def __repr__(self):
        if _Debug:
            lg.args(_DebugLevel*2, latest_as_string=self.latest_as_string)
        return '{%s%s}' % ('' if self.is_latest() else '*', self.latest_as_string)

    def __str__(self):
        if _Debug:
            lg.args(_DebugLevel*2, latest_as_string=self.latest_as_string)
        return self.latest_as_string

    def __bytes__(self):
        if _Debug:
            lg.args(_DebugLevel*2, latest=self.latest)
        return self.latest

    def __bool__(self):
        if _Debug:
            lg.args(_DebugLevel*2, latest=self.latest)
        return bool(self.latest)

    def __len__(self):
        if _Debug:
            lg.args(_DebugLevel*2, latest=self.latest)
        return len(self.latest)

    def refresh(self, replace_original=True):
        _latest, _latest_revision = get_latest_revision(self.current)
        if self.latest and self.latest == _latest:
            if _Debug:
                lg.args(_DebugLevel, latest=self.latest_as_string, refreshed=False)
            return False
        self.latest = _latest
        self.latest_revision = _latest_revision
        if not self.latest:
            self.latest = self.current
            self.latest_revision = -1
        self.latest_as_string = strng.to_text(self.latest)
        latest_username, latest_host = idurl_to_id(self.latest, by_parts=True)
        if latest_username != self.username:
            caller_code = sys._getframe().f_back.f_code
            caller_method = caller_code.co_name
            caller_modul = os.path.basename(caller_code.co_filename).replace('.py', '')
            if caller_method.count('lambda') or caller_method == 'field':
                caller_method = sys._getframe(1).f_back.f_code.co_name
            exc = ValueError('while refreshing tried to modify username of the identity %r -> %r' % (self.current, self.latest))
            lg.exc(msg='called from %s.%s()' % (caller_modul, caller_method), exc_value=exc)
            raise exc
        self.latest_host = latest_host
        self.latest_id = '%s@%s' % (self.username, self.latest_host)
        if replace_original:
            self.current = self.latest
            self.current_as_string = self.latest_as_string
            self.current_id = self.latest_id
        if _Debug:
            lg.args(_DebugLevel, latest=self.latest_as_string, current=self.current_as_string, refreshed=True)
        return True

    def strip(self):
        if _Debug:
            lg.args(_DebugLevel*2, latest=self.latest_as_string)
        return self.latest

    def original(self):
        if _Debug:
            lg.args(_DebugLevel*2, current=self.current, latest=self.latest)
        return self.current

    def original_id(self):
        if _Debug:
            lg.args(_DebugLevel*2, current=self.current_id, latest=self.latest_id)
        return self.current_id

    def is_latest(self):
        if _Debug:
            lg.args(_DebugLevel*2, current=self.current, latest=self.latest)
        return self.latest and self.current == self.latest

    def to_id(self):
        if _Debug:
            lg.args(_DebugLevel*2, latest_id=self.latest_id)
        return self.latest_id

    def to_text(self):
        if _Debug:
            lg.args(_DebugLevel*2, latest=self.latest_as_string)
        return self.latest_as_string

    def to_bin(self):
        if _Debug:
            lg.args(_DebugLevel*2, latest=self.latest)
        return self.latest

    def to_original(self):
        if _Debug:
            lg.args(_DebugLevel*2, current=self.current, latest=self.latest)
        return self.current

    def to_public_key(self, raise_error=True):
        global _KnownIDURLs
        if _Debug:
            lg.args(_DebugLevel*2, latest=self.latest, current=self.current)
        if not self.current:
            return b''
        if self.current not in _KnownIDURLs:
            if not raise_error:
                return None
            caller_code = sys._getframe().f_back.f_code
            caller_method = caller_code.co_name
            caller_modul = os.path.basename(caller_code.co_filename).replace('.py', '')
            if caller_method.count('lambda') or caller_method.startswith('_'):
                caller_method = sys._getframe(1).f_back.f_code.co_name
            exc = KeyError('unknown idurl %r in %s.%s' % (self.current, caller_modul, caller_method))
            lg.exc(msg='called from %s.%s()' % (caller_modul, caller_method), exc_value=exc)
            if _Debug:
                lg.out(_DebugLevel, traceback.format_list(traceback.format_stack()))
            raise exc
        pub_key = _KnownIDURLs[self.current]
        return pub_key

    def unique_name(self, raise_error=True):
        if not self._unique_name:
            self._unique_name = '{}_{}'.format(
                self.username,
                strng.to_text(hashes.sha1(self.to_public_key(raise_error=raise_error), hexdigest=True)),
            )
        return self._unique_name



#!/usr/bin/python
# help.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (help.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: help.

A methods to just store text constants, used to print command-line
instructions.
"""


def usage_text():
    return '''usage: bitdust [options] [command] [arguments]

Commands:
  install
  start
  detach
  restart
  stop
  show
  alias
  id
  idurl
  identity create <username> [private key size]
  identity backup [private key file path]
  identity restore <private key file path> [IDURL]
  identity erase
  identity show
  identity server [start | stop]
  key list
  key create <key id>
  key delete <key id>
  key get [key_id]
  key copy [key id]
  key backup <key id> <filename>
  get <option>
  set <option> [value]
  set list
  file sync
  file list
  file create <remote path>
  file delete <remote path>
  dir create <remote path>
  dir delete <remote path>
  file upload <local path> <remote path>
  file download <remote path> <local path>
  file cancel <upload | download> <remote path>
  file progress [upload | download]
  supplier list
  supplier replace <IDURL or position>
  supplier change <IDURL or position> <new IDURL>
  supplier ping
  customer list
  customer remove <IDURL>
  customer ping
  storage
  automat list
  service list
  service <service name>
  ping <IDURL>
  dhtseed
  chat
  chat send <IDURL> "<text message>"
  api <method> [params]
  version
  help
  usage
'''


#------------------------------------------------------------------------------


def help_text():
    return '''usage: bitdust [options] [command] [arguments]

Commands:
  install               create virtual environment and deploy required
                        Python packages in ~/.bitdust/venv/

  [start]               start main BitDust process

  detach                start BitDust in as a daemon process

  restart               restart BitDust

  stop                  stop BitDust

  show                  start BitDust and show the main window

  alias                 helper to create a binary command-alias in OS,
                        you can put it in /usr/local/bin/bitdust

  id                    print your global ID in BitDust network

  idurl                 print your global IDURL: global address of
                        your identity file in the Internet

  identity create <nickname> [private key size]
                        generate a new private key and
                        new identity file for you
                        key size can be 1024, 2048 or 4096

  identity backup [private key file path]
                        backup your IDURL and private key to the file,
                        to be able to restore your data in case of lost
                        ATTENTION! always keep a backup copy of your
                        master key in safe place, this is the only
                        possible way to recover access to your lost data

  identity restore <private key file path> [IDURL]
                        recover existing identity file
                        with your private key file

  identity erase        delete local identity from this machine

  identity show         displays your current identity file content

  identity server start|stop
                        start/stop stand alone identity server on that machine

  key list              list details for known private keys

  key create <key_id> [size]
                        generate a new private key with given ID

  key delete <key_id>
                        erase given private key
                        WARNING!!! all data encrypted with that key will be lost

  key get [key_id]      prints private key details to console
                        WARNING!!! never publish your "master" key

  key copy [key_id]     copy given private key to clipboard, use Ctrl+V to paste it
                        WARNING!!! never publish your "master" key

  key backup <key_id> <filename>
                        copy private key into file

  get <option>          print current value for given program setting

  set <option> [value]  assign a new value for program setting

  set list              print all available options and current values

  file sync             ping all suppliers and recheck/restart uploads

  file list             show a list of known files and folders on remote peers

  file create <remote path>
                        creates "virtual" remote file in the catalog,
                        "remote path" have such format in all commands:
                        {key_alias}${user_name}@{id_host}:{remote_path}

  file delete <remote path>
                        remove give file from the catalog and remote suppliers

  dir create <remote path>
                        create "virtual" remote folder in the catalog

  dir delete <remote path>
                        erase folder and all sub-folders from the catalog
                        and remote suppliers

  file upload <local path> <remote path>
                        start uploading a file or folder to remote suppliers,
                        first you need to create "remote path" in the catalog

  file download <remote path> [local path]
                        download file or folder from remote peers to local drive
                        save to current folder if "local path" was not set

  file cancel <upload | download> <remote path>
                        abort currently running file upload or download process

  file progress <upload | download>
                        show a list of currently running uploads or/and downloads

  supplier list         show list of your suppliers
                        nodes who stores your data on own machines

  supplier replace <IDURL or position>
                        execute a fire/hire process for given supplier,
                        another random node will replace him

  supplier change <IDURL or position> <new IDURL>
                        replace a supplier by given node

  supplier ping         send Identity packet to all suppliers
                        and wait Ack packets to check their statuses

  customer list         show list of your customers
                        nodes who stores own data on your machine

  customer remove <IDURL>
                        reject supporting given customer and
                        remove all local files stored for him

  customer ping         send Identity packet to all customers
                        and wait Ack packets to check their statuses

  storage               show donated/needed storage statistic

  automat list          list all running state machines and current states

  service list          list all registered services

  service <service name>
                        print detailed info for given service

  ping <IDURL>          send Identity packet to this node
                        and wait Ack packet to check his status

  dhtseed               start stand alone Distributed Hash Table seed node

  chat                  start a chat session, send/listen text
                        messages from other users

  chat send <IDURL> "<text message>"
                        send a single text message to remote user

  api <method> [params] execute API method and return JSON response

  version               display current software version

  help                  print a detailed info about command line usage

  usage                 print a brief list of available commands

'''


#------------------------------------------------------------------------------


def schedule_format():
    return '''
Schedule compact format:
[mode].[interval].[time].[details]

mode:
  n-none, h-hourly, d-daily, w-weekly, m-monthly, c-continuously

interval:
  just a number - how often to restart the task, default is 1

time:
  [hour]:[minute]

details:
  for weeks: Mon Tue Wed Thu Fri Sat Sun
  for months: Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec

some examples:
  none                    no schedule
  hourly.3                each 3 hours
  daily.4.10:15.          every 4th day at 10:15
  w.1.3:00.MonSat         every Monday and Saturday at 3:00 in the night
  weekly.4.18:45.MonTueWedThuFriSatSun
                          every day in each 4th week in 18:45
  m.5.12:34.JanJul        5th Jan and 5th July at 12:34
  c.300                   every 300 seconds (10 minutes)
'''


def settings_help():
    return '''set [option] [value]

examples:
  set donated 4GB                          set donated space
  set needed                               print your needed space size
  set services/backups/max-copies 4        set number of backup copies for every folder
  set services/customer/suppliers-number   print number of your suppliers
  set list                                 list all available options

'''



#!/usr/bin/python
# service_miner.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_miner.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_miner
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return MinerService()


class MinerService(LocalService):

    service_name = 'service_miner'
    config_path = 'services/miner/enabled'

    def dependent_on(self):
        return [
            'service_nodes_lookup',
            # 'service_blockchain',
        ]

    def installed(self):
        # TODO: to be continue...
        return False

    def start(self):
        from coins import coins_miner
        coins_miner.A('init')
        coins_miner.A('start')
        return True

    def stop(self):
        from coins import coins_miner
        coins_miner.A('stop')
        coins_miner.A('shutdown')
        return True

    def request(self, json_payload, newpacket, info):
        # TODO: work in progress
        # TODO: we can add some limit for number of connections here
        from bitdust.p2p import p2p_service
        return p2p_service.SendAck(newpacket, 'accepted')

    def cancel(self, json_payload, request, info):
        # TODO: work in progress
        from bitdust.p2p import p2p_service
        return p2p_service.SendAck(request, 'accepted')



import os
import math
import threading
import time
import socks
import hashlib
import random

#------------------------------------------------------------------------------

from twisted.internet import reactor

#------------------------------------------------------------------------------

from bitdust_forks.Bismuth import mining_heavy3  # @UnresolvedImport
from bitdust_forks.Bismuth import connections  # @UnresolvedImport

from bitdust.logs import lg
from bitdust.main import settings
from bitdust.main import config

from bitdust.blockchain import known_bismuth_nodes
from bitdust.blockchain import bismuth_wallet

from bitdust.services import driver

from bitdust.userid import my_id

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

_DataDirPath = None
_MinerWalletAddress = None
_MinerName = None
_OwnCoinsLastTime = None
_WantMoreCoins = False
_MiningIsOn = False
_MiningPoolHost = None
_MiningPoolPort = None

#------------------------------------------------------------------------------

mining_threads = 1
nonce_time = 10
max_diff = 150
hashcount = 20000

#------------------------------------------------------------------------------


def init():
    global _DataDirPath
    global _MinerName
    global _MinerWalletAddress
    global _MiningPoolHost
    global _MiningPoolPort
    _DataDirPath = settings.ServiceDir('bismuth_blockchain')
    _MinerWalletAddress = bismuth_wallet.my_wallet_address()
    _MinerName = my_id.getIDName()
    if driver.is_enabled('service_bismuth_pool'):
        _MiningPoolHost = config.conf().getString('services/bismuth-pool/host', '127.0.0.1')
        _MiningPoolPort = config.conf().getInt('services/bismuth-pool/tcp-port', 18525)
    check_start_mining_later(delay=0)
    if _Debug:
        lg.args(_DebugLevel, wallet_address=_MinerWalletAddress, miner_name=_MinerName)
    return True


def shutdown():
    return True


#------------------------------------------------------------------------------


def get_random_mining_pool_host_port():
    one_item = random.choice(list(known_bismuth_nodes.mining_pools_by_host().items()))
    return one_item[0], one_item[1]


#------------------------------------------------------------------------------


def check_start_mining_later(delay=30):
    # TODO: use random delay
    reactor.callLater(delay, check_start_mining)  # @UndefinedVariable


def check_start_mining():
    global _OwnCoinsLastTime
    global _WantMoreCoins
    global _MiningIsOn

    cur_balance = bismuth_wallet.my_balance()

    lg.info('my wallet address is %s and current balance is %s' % (bismuth_wallet.my_wallet_address(), cur_balance))
    if _Debug:
        lg.args(_DebugLevel, MiningIsOn=_MiningIsOn, WantMoreCoins=_WantMoreCoins, OwnCoinsLastTime=_OwnCoinsLastTime)

    if cur_balance == 'N/A':
        reactor.callLater(10, check_start_mining)  # @UndefinedVariable
        return

    if _WantMoreCoins and (not _OwnCoinsLastTime or (time.time() - _OwnCoinsLastTime > 60)):
        _OwnCoinsLastTime = time.time()
        _WantMoreCoins = False
        run(needed_coins=1)
    else:
        run(needed_coins=0)


def run(needed_coins):
    global _MinerWalletAddress
    global _MinerName
    global _DataDirPath
    global _MiningPoolHost
    global _MiningPoolPort
    global _MiningIsOn

    if not _MiningPoolPort or not _MiningPoolHost:
        _MiningPoolHost, _MiningPoolPort = get_random_mining_pool_host_port()

    if _Debug:
        lg.args(_DebugLevel, needed_coins=needed_coins, mining_pool_host=_MiningPoolHost, mining_pool_port=_MiningPoolPort)
    _MiningIsOn = True

    miner_th = threading.Thread(target=miner_thread, args=(
        needed_coins,
        _MiningPoolHost,
        _MiningPoolPort,
        _MinerWalletAddress,
        _MinerName,
        _DataDirPath,
    ))
    miner_th.daemon = True
    miner_th.start()


def miner_thread(needed_coins, mining_pool_host, mining_pool_port, miner_address, miner_name, data_dir_path):
    global _MiningIsOn

    if not mining_heavy3.RND_LEN or not mining_heavy3.MMAP:
        mining_heavy3.mining_open(os.path.join(data_dir_path, 'heavy3a.bin'))

    mined_coins = 0
    delay = 0
    while True:
        if not _MiningIsOn:
            break

        if needed_coins and mined_coins >= needed_coins:
            if _Debug:
                lg.dbg(_DebugLevel, 'successfully mined %d coins, finishing' % mined_coins)
            break

        # print('Miner: Starting active mining, progress so far: %d/%d, requesting work from the mining pool: %s:%s' % (
        #     mined_coins, needed_coins, mining_pool_host, int(mining_pool_port)
        # ))

        try:
            s = socks.socksocket()
            s.connect((mining_pool_host, int(mining_pool_port)))
            connections.send(s, 'getwork', 10)
            work_pack = connections.receive(s, 10)
            mempool_size = (work_pack[-1][0])
            if not needed_coins and not mempool_size:
                # if _Debug:
                #     lg.args(_DebugLevel, needed_coins=needed_coins, mempool_size=mempool_size)
                # print('Miner: Mempool is empty, skip mining')
                delay = 30
                break
                # mining_heavy3.mining_close()
                # time.sleep(30)
                # continue
                # return

            db_block_hash = work_pack[-1][1]
            diff = int(work_pack[-1][2])
            pool_address = work_pack[-1][3]
            netdiff = int(work_pack[-1][4])
            s.close()

            diff_hex = math.floor((diff/8) - 1)
            mining_condition = db_block_hash[0:diff_hex]

            h1 = 1

            # print('Miner instance started', miner_address, pool_address, db_block_hash, diff, mining_condition, netdiff)

            try:
                tries = 0
                try_arr = [('%0x' % random.getrandbits(32)) for _ in range(nonce_time*hashcount)]
                address = pool_address
                success = False

                while not success:
                    if not _MiningIsOn:
                        break

                    if needed_coins and mined_coins >= needed_coins:
                        break

                    try:
                        t1 = time.time()
                        tries = tries + 1
                        # generate the "address" of a random backyard that we will sample in this try
                        seed = ('%0x' % random.getrandbits(128 - 32))
                        # this part won't change, so concat once only
                        prefix = pool_address + seed
                        # This is where the actual hashing takes place
                        # possibles = [nonce for nonce in try_arr if mining_condition in (sha224((prefix + nonce + db_block_hash).encode("utf-8")).hexdigest())]
                        possibles = [nonce for nonce in try_arr if mining_condition in (mining_heavy3.anneal3(mining_heavy3.MMAP, int.from_bytes(hashlib.sha224((prefix + nonce + db_block_hash).encode('utf-8')).digest(), 'big')))]
                        # hash rate calculation
                        try:
                            t2 = time.time()
                            h1 = int(((nonce_time*hashcount)/(t2 - t1))/1000)
                        except Exception as e:
                            lg.exc()
                            h1 = 1
                        if possibles:
                            for nonce in possibles:
                                # add the seed back to get a full 128 bits nonce
                                nonce = seed + nonce
                                xdiffx = mining_heavy3.diffme_heavy3(address, nonce, db_block_hash)
                                if xdiffx < diff:
                                    pass

                                else:
                                    # print('Miner: Solved work with difficulty {} in {} cycles - YAY!'.format(xdiffx, tries))
                                    # wname = '{}{}'.format(miner_name, 0)
                                    # print('Miner: {} running at {} kh/s'.format(wname, str(h1)))
                                    block_send = []
                                    del block_send[:]  # empty
                                    block_timestamp = '%.2f' % time.time()
                                    block_send.append((block_timestamp, nonce, db_block_hash, netdiff, xdiffx, 0, miner_name, 1, str(1)))
                                    # print('Miner: Sending solution: {}'.format(block_send))
                                    tries = 0
                                    # submit mined nonce to pool
                                    try:
                                        s1 = socks.socksocket()
                                        s1.connect((mining_pool_host, int(mining_pool_port)))  # connect to pool
                                        # print('Miner: connected to pool, proceeding to submit solution miner_address=%s' % miner_address)
                                        connections.send(s1, 'block', 10)
                                        connections.send(s1, miner_address, 10)
                                        connections.send(s1, block_send, 10)
                                        time.sleep(0.2)
                                        s1.close()
                                        mined_coins += 1
                                        success = True
                                        # print('Miner: solution submitted to pool', mined_coins)
                                        if _Debug:
                                            lg.args(_DebugLevel, xdiffx=xdiffx, tries=tries, hash_rate=h1, mined_coins=mined_coins)
                                        break

                                    except Exception as e:
                                        lg.exc()
                                        # print('Miner: Could not submit solution to pool')

                    except Exception as e:
                        # traceback.print_exc()
                        lg.exc()
                        time.sleep(0.1)
                        raise

            except Exception as e:
                lg.exc()

        except Exception as e:
            lg.warn('unable to connect to mining pool: %r' % e)
            # print('Miner: Unable to connect to pool, check your connection or IP settings', e)
            time.sleep(5)

    mining_heavy3.mining_close()
    _MiningIsOn = False
    # print('Miner: thread finished, mined_coins=%d, needed_coins=%d' % (mined_coins, needed_coins, ))
    reactor.callFromThread(check_start_mining_later, delay=delay)  # @UndefinedVariable


#------------------------------------------------------------------------------

bin_format_dict = dict((x, format(ord(x), '8b').replace(' ', '0')) for x in '0123456789abcdef')


def bin_convert(string):
    return ''.join(bin_format_dict[x] for x in string)


def bin_convert_orig(string):
    return ''.join(format(ord(x), '8b').replace(' ', '0') for x in string)


def diffme(pool_address, nonce, db_block_hash):
    diff = 60
    diff_result = 0
    mining_hash = bin_convert(hashlib.sha224((pool_address + nonce + db_block_hash).encode('utf-8')).hexdigest())
    mining_condition = bin_convert(db_block_hash)
    while mining_condition[:diff] in mining_hash:
        diff_result = diff
        diff += 1
    return diff_result



#!/usr/bin/env python
# udp_node.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (udp_node.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: udp_node.

.. role:: red
BitDust udp_node() Automat


EVENTS:
    * :red:`connect`
    * :red:`datagram-received`
    * :red:`dht-read-result`
    * :red:`dht-write-failed`
    * :red:`dht-write-success`
    * :red:`disconnected`
    * :red:`go-offline`
    * :red:`go-online`
    * :red:`stun-failed`
    * :red:`stun-success`
    * :red:`timer-10sec`
    * :red:`timer-1sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

import time

from twisted.internet import reactor  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import udp
from bitdust.lib import strng
from bitdust.lib import misc

from bitdust.main import settings

from bitdust.dht import dht_service

from bitdust.services import driver

from bitdust.transport.udp import udp_connector
from bitdust.transport.udp import udp_session

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 18

#------------------------------------------------------------------------------

_UDPNode = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _UDPNode
    if _UDPNode is None:
        # set automat name and starting state here
        _UDPNode = UDPNode('udp_node', 'AT_STARTUP', 24)
    if event is not None:
        _UDPNode.automat(event, *args, **kwargs)
    return _UDPNode


def Destroy():
    """
    Destroy udp_node() automat and remove its instance from memory.
    """
    global _UDPNode
    if _UDPNode is None:
        return
    _UDPNode.destroy()
    del _UDPNode
    _UDPNode = None


class UDPNode(automat.Automat):
    """
    This class implements all the functionality of the ``udp_node()`` state
    machine.
    """

    fast = True

    timers = {
        'timer-1sec': (1.0, ['LISTEN']),
        'timer-10sec': (10.0, ['LISTEN']),
    }

    def init(self):
        """
        Method to initialize additional variables and flags at creation of the
        state machine.
        """
        self.listen_port = None
        self.my_id = None
        self.my_address = None
        self.options = {}
        if driver.is_on('service_my_ip_port'):
            from bitdust.stun import stun_client
            self.my_address = stun_client.A().getMyExternalAddress()
        self.notified = False
        self.IncomingPosition = -1

    def A(self, event, *args, **kwargs):
        #---LISTEN---
        if self.state == 'LISTEN':
            if event == 'datagram-received' and self.isPacketValid(*args, **kwargs) and not self.isStun(*args, **kwargs) and not self.isKnownPeer(*args, **kwargs):
                self.doStartNewSession(*args, **kwargs)
            elif event == 'go-offline':
                self.state = 'DISCONNECTING'
                self.doDisconnect(*args, **kwargs)
            elif event == 'connect' and self.isKnowMyAddress(*args, **kwargs) and not self.isKnownUser(*args, **kwargs):
                self.doStartNewConnector(*args, **kwargs)
            elif event == 'timer-10sec' and not self.isKnowMyAddress(*args, **kwargs):
                self.state = 'STUN'
                self.doStartStunClient(*args, **kwargs)
            elif event == 'timer-10sec' and self.isKnowMyAddress(*args, **kwargs):
                self.state = 'WRITE_MY_IP'
                self.doDHTWtiteMyAddress(*args, **kwargs)
            elif event == 'dht-read-result':
                self.doCheckAndStartNewSession(*args, **kwargs)
                self.doDHTRemoveMyIncoming(*args, **kwargs)
                self.doNotifyConnected(*args, **kwargs)
            elif event == 'connect' and not self.isKnowMyAddress(*args, **kwargs):
                self.state = 'STUN'
                self.doStartStunClient(*args, **kwargs)
            elif event == 'timer-1sec':
                self.doDHTReadNextIncoming(*args, **kwargs)
        #---AT_STARTUP---
        elif self.state == 'AT_STARTUP':
            if event == 'go-online' and not self.isKnowMyAddress(*args, **kwargs):
                self.state = 'STUN'
                self.GoOn = False
                self.doInit(*args, **kwargs)
                self.doStartStunClient(*args, **kwargs)
            elif event == 'go-online' and self.isKnowMyAddress(*args, **kwargs):
                self.state = 'WRITE_MY_IP'
                self.GoOn = False
                self.doInit(*args, **kwargs)
                self.doDHTWtiteMyAddress(*args, **kwargs)
        #---STUN---
        elif self.state == 'STUN':
            if event == 'stun-success':
                self.state = 'WRITE_MY_IP'
                self.doUpdateMyAddress(*args, **kwargs)
                self.doDHTWtiteMyAddress(*args, **kwargs)
            elif event == 'go-offline':
                self.state = 'DISCONNECTING'
                self.doDisconnect(*args, **kwargs)
            elif event == 'stun-failed':
                self.state = 'OFFLINE'
                self.doUpdateMyAddress(*args, **kwargs)
                self.doNotifyFailed(*args, **kwargs)
        #---OFFLINE---
        elif self.state == 'OFFLINE':
            if event == 'go-online':
                self.state = 'STUN'
                self.doStartStunClient(*args, **kwargs)
            elif event == 'go-offline':
                self.doNotifyDisconnected(*args, **kwargs)
        #---WRITE_MY_IP---
        elif self.state == 'WRITE_MY_IP':
            if event == 'go-offline':
                self.state = 'DISCONNECTING'
                self.doDisconnect(*args, **kwargs)
            elif event == 'connect' and not self.isKnowMyAddress(*args, **kwargs):
                self.state = 'STUN'
                self.doStartStunClient(*args, **kwargs)
            elif event == 'connect' and self.isKnowMyAddress(*args, **kwargs) and not self.isKnownUser(*args, **kwargs):
                self.doStartNewConnector(*args, **kwargs)
            elif event == 'dht-write-failed':
                self.state = 'OFFLINE'
                self.doNotifyFailed(*args, **kwargs)
            elif event == 'dht-write-success':
                self.state = 'LISTEN'
                self.doDHTReadNextIncoming(*args, **kwargs)
        #---DISCONNECTING---
        elif self.state == 'DISCONNECTING':
            if event == 'go-online':
                self.GoOn = True
            elif event == 'disconnected' and not self.GoOn:
                self.state = 'OFFLINE'
                self.doNotifyDisconnected(*args, **kwargs)
            elif event == 'disconnected' and self.GoOn:
                self.state = 'STUN'
                self.GoOn = False
                self.doNotifyDisconnected(*args, **kwargs)
                self.doStartStunClient(*args, **kwargs)
        return None

    def isKnownPeer(self, *args, **kwargs):
        """
        Condition method.
        """
        try:
            datagram, address = args[0]
            command, payload = datagram
        except:
            lg.exc()
            return False
        if driver.is_on('service_my_ip_port'):
            from bitdust.stun import stun_client
            if address in stun_client.A().stun_servers:
                return True
        active_sessions = udp_session.get(address)
        return len(active_sessions) > 0

    def isKnownUser(self, *args, **kwargs):
        """
        Condition method.
        """
        user_id = args[0]
        if udp_session.get_by_peer_id(user_id):
            return True
        if udp_connector.get(user_id) is not None:
            return True
        if _Debug:
            lg.out(_DebugLevel, 'udp_node.isKnownUser %s not found in %s' % (user_id, list(udp_session.sessions_by_peer_id().keys())))
        return False

    def isKnowMyAddress(self, *args, **kwargs):
        """
        Condition method.
        """
        return self.my_address is not None

    def isPacketValid(self, *args, **kwargs):
        """
        Condition method.
        """
        try:
            datagram, address = args[0]
            command, payload = datagram
        except:
            return False
        return True

    def isStun(self, *args, **kwargs):
        """
        Condition method.
        """
        command = args[0][0][0]
        return command == udp.CMD_STUN

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport.udp import udp_interface
        from bitdust.transport.udp import udp_stream
        self.options = args[0]
        self.my_idurl = strng.to_text(self.options['idurl'])
        self.listen_port = int(self.options['udp_port'])
        self.my_id = udp_interface.idurl_to_id(self.my_idurl)
        udp.proto(self.listen_port).add_callback(self._datagram_received)
        bandoutlimit = settings.getBandOutLimit()
        bandinlimit = settings.getBandInLimit()
        udp_stream.set_global_output_limit_bytes_per_sec(bandoutlimit)
        udp_stream.set_global_input_limit_bytes_per_sec(bandinlimit)
        reactor.callLater(0, udp_session.process_sessions)  # @UndefinedVariable
        reactor.callLater(0, udp_stream.process_streams)  # @UndefinedVariable

    def doStartStunClient(self, *args, **kwargs):
        """
        Action method.
        """
        if driver.is_on('service_my_ip_port'):
            from bitdust.stun import stun_client
            stun_client.A('start', self._stun_finished)
        else:
            self.automat('stun-success', ('unknown', misc.readExternalIP() or '127.0.0.1', settings.getUDPPort()))

    def doStartNewConnector(self, *args, **kwargs):
        """
        Action method.
        """
        c = udp_connector.create(self, *args, **kwargs)
        c.automat('start', (self.listen_port, self.my_id, self.my_address))

    def doStartNewSession(self, *args, **kwargs):
        """
        Action method.
        """
        try:
            datagram, address = args[0]
            command, payload = datagram
        except:
            lg.exc()
            return
        if _Debug:
            lg.out(_DebugLevel, 'udp_node.doStartNewSession wants to start a new session with UNKNOWN peer at %s' % str(address))
        s = udp_session.create(self, address)
        s.automat('init')
        s.automat('datagram-received', *args, **kwargs)

    def doCheckAndStartNewSession(self, *args, **kwargs):
        """
        Action method.
        """
        if self.my_address is None:
            if _Debug:
                lg.out(_DebugLevel, 'dp_node.doCheckAndStartNewSession SKIP because my_address is None')
            return
        incoming_str = args[0]
        if incoming_str is None:
            return
        try:
            incoming_user_id, incoming_user_address, _ = incoming_str.split(b' ')
            incoming_user_address = incoming_user_address.split(b':')
            incoming_user_address[1] = int(incoming_user_address[1])
            incoming_user_address = tuple(incoming_user_address)
        except:
            if _Debug:
                lg.out(_DebugLevel, '%r' % incoming_str)
            lg.exc()
            return
        active_sessions = udp_session.get(incoming_user_address)
        if active_sessions:
            if _Debug:
                lg.out(_DebugLevel, 'udp_node.doCheckAndStartNewSessions SKIP because found existing by address %s : %s' % (incoming_user_address, active_sessions))
            return
        active_sessions = udp_session.get_by_peer_id(incoming_user_id)
        if active_sessions:
            if _Debug:
                lg.out(_DebugLevel, 'udp_node.doCheckAndStartNewSession SKIP because found existing by peer id %s : %s' % (incoming_user_id, active_sessions))
            return
        if _Debug:
            lg.out(_DebugLevel, 'udp_node.doCheckAndStartNewSession wants to start a new session with incoming peer %s at %s' % (incoming_user_id, incoming_user_address))
        s = udp_session.create(self, incoming_user_address, incoming_user_id)
        s.automat('init')

    def doUpdateMyAddress(self, *args, **kwargs):
        """
        Action method.
        """
        try:
            typ, new_ip, new_port = args[0]
            new_addr = (new_ip, new_port)
        except:
            lg.exc()
            return
        if _Debug:
            lg.out(4, 'udp_node.doUpdateMyAddress typ=[%s]' % typ)
            if self.my_address:
                lg.out(4, '    old=%s new=%s' % (str(self.my_address), str(new_addr)))
            else:
                lg.out(4, '    new=%s' % str(new_addr))
        self.my_address = new_addr

    def doDHTReadNextIncoming(self, *args, **kwargs):
        """
        Action method.
        """
        self.IncomingPosition += 1
        if self.IncomingPosition >= 10:
            self.IncomingPosition = 0
        key = self.my_id + ':incoming' + str(self.IncomingPosition)
        if _Debug:
            lg.out(_DebugLevel, 'udp_node.doDHTReadNextIncoming  key=%s' % key)
        d = dht_service.get_value(key)
        d.addCallback(self._got_my_incoming, key, self.IncomingPosition)
        d.addErrback(self._failed_my_incoming, key, self.IncomingPosition)

    def doDHTRemoveMyIncoming(self, *args, **kwargs):
        """
        Action method.
        """
        if args and args[0]:
            key = self.my_id + ':incoming' + str(self.IncomingPosition)
            dht_service.delete_key(key)

    def doDHTWtiteMyAddress(self, *args, **kwargs):
        """
        Action method.
        """
        d = dht_service.set_value(
            self.my_id + ':address',
            '%s:%d' % (strng.to_bin(self.my_address[0]), self.my_address[1]),
            age=int(time.time()),
        )
        d.addCallback(self._wrote_my_address)
        d.addErrback(lambda x: self.automat('dht-write-failed'))

    def doDisconnect(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport.udp import udp_stream
        lg.out(12, 'udp_node.doDisconnect going to close %d sessions and %d connectors' % (len(list(udp_session.sessions().values())), len(list(udp_connector.connectors().values()))))
        udp_stream.stop_process_streams()
        udp_session.stop_process_sessions()
        for s in udp_session.sessions().values():
            if _Debug:
                lg.out(_DebugLevel, 'udp_node.doShutdown sends "shutdown" to %s' % s)
            s.automat('shutdown')
        for c in udp_connector.connectors().values():
            c.automat('abort')
        self.automat('disconnected')

    def doNotifyDisconnected(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport.udp import udp_interface
        self.notified = False
        udp_interface.interface_disconnected(args[0])

    def doNotifyConnected(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport.udp import udp_interface
        if not self.notified:
            udp_interface.interface_receiving_started(self.my_id, self.options)
            self.notified = True
            if _Debug:
                lg.out(4, 'udp_node.doNotifyConnected my host is %s' % self.my_id)

    def doNotifyFailed(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport.udp import udp_interface
        udp_interface.interface_receiving_failed('state is %s' % self.state)

    def _datagram_received(self, datagram, address):
        """
        """
        # lg.out(18, '-> [%s] (%d bytes) from %s' % (command, len(payload), str(address)))
        active_sessions = udp_session.get(address)
        if active_sessions:
            for s in active_sessions:
                s.automat('datagram-received', (datagram, address))
        self.automat('datagram-received', (datagram, address))
        return False

    def _stun_finished(self, result, typ, ip, details):
        if result == 'stun-success' and typ == 'symmetric':
            result = 'stun-failed'
        self.automat(result, (typ, ip, details))

    def _got_my_address(self, value, key):
        if not isinstance(value, dict):
            lg.warn('can not read my address')
            self.automat('dht-write-failed')
            return
        try:
            addr = value[dht_service.key_to_hash(key)].strip('\n').strip()
        except:
            if _Debug:
                lg.out(4, 'udp_node._got_my_address ERROR   wrong key in response: %r' % value)
                lg.exc()
            self.automat('dht-write-failed')
            return
        if addr != '%s:%d' % (self.my_address[0], self.my_address[1]):
            if _Debug:
                lg.out(4, 'udp_node._got_my_address ERROR   value not fit: %r' % value)
            self.automat('dht-write-failed')
            return
        self.automat('dht-write-success')

    def _wrote_my_address(self, nodes):
        if len(nodes) == 0:
            self.automat('dht-write-failed')
            return
        key = self.my_id + ':address'
        d = dht_service.get_value(key)
        d.addCallback(self._got_my_address, key)
        d.addErrback(lambda x: self.automat('dht-write-failed'))

    def _got_my_incoming(self, value, key, position):
        if not isinstance(value, dict):
            if _Debug:
                lg.out(_DebugLevel, 'udp_node._got_my_incoming no incoming at position: %d' % position)
            self.automat('dht-read-result', None)
            return
        try:
            myincoming = value[dht_service.key_to_hash(key)]
        except:
            if _Debug:
                lg.out(_DebugLevel, 'udp_node._got_my_incoming ERROR reading my incoming at position: %d\n%r' % (position, value))
            self.automat('dht-read-result', None)
            return
        if _Debug:
            lg.out(_DebugLevel, 'udp_node._got_my_incoming found one: %r' % myincoming)
        self.automat('dht-read-result', myincoming)

    def _failed_my_incoming(self, err, key, position):
        if _Debug:
            lg.out(_DebugLevel, 'udp_node._got_my_incoming incoming empty: %s' % str(position))
        self.automat('dht-read-result', None)



#!/usr/bin/python
# http_interface.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (http_interface.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: http_interface.

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False

#------------------------------------------------------------------------------

import six
import sys

try:
    pass
except:
    sys.exit('Error initializing twisted.internet.reactor in http_interface.py')

from twisted.web import xmlrpc
from twisted.internet.defer import succeed, fail

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.main import settings

from bitdust.lib import misc
from bitdust.lib import strng
from bitdust.lib import net_misc

from bitdust.transport.http import http_node

#------------------------------------------------------------------------------

_GateProxy = None

#------------------------------------------------------------------------------


def proxy():
    global _GateProxy
    return _GateProxy


#------------------------------------------------------------------------------


class GateInterface():
    def init(self, xml_rpc_url_or_object):
        """
        """
        if _Debug:
            lg.out(4, 'http_interface.init')
        if not proxy():
            global _GateProxy
            if isinstance(xml_rpc_url_or_object, six.string_types):
                _GateProxy = xmlrpc.Proxy(xml_rpc_url_or_object, allowNone=True)
            else:
                _GateProxy = xml_rpc_url_or_object
        proxy().callRemote('transport_initialized', 'http')
        return True

    def shutdown(self):
        """
        """
        if _Debug:
            lg.out(4, 'http_interface.shutdown')
        if proxy():
            global _GateProxy
            del _GateProxy
            _GateProxy = None
        return True

    def connect(self, options):
        """
        """
        if _Debug:
            lg.out(4, 'http_interface.connect %s' % str(options))
        if settings.enableHTTPreceiving():
            http_node.start_receiving()
        else:
            lg.warn('transport_http receiving is disabled')
            interface_receiving_failed()
            return False
        if settings.enableHTTPsending():
            http_node.start_sending(port=options['http_port'])
        return succeed(True)

    def disconnect(self):
        """
        """
        if _Debug:
            lg.out(4, 'http_interface.disconnect')
        http_node.stop_sending()
        http_node.stop_receiving()
        return succeed(True)

    def build_contacts(self, id_obj):
        """
        """
        result = []
        nowip = strng.to_bin(misc.readExternalIP())
        result.append(b'http://%s:%d' % (nowip, settings.getHTTPPort()))
        if _Debug:
            lg.out(4, 'http_interface.build_contacts : %s' % result)
        return result

    def verify_contacts(self, id_obj):
        """
        """
        nowip = strng.to_bin(misc.readExternalIP())
        http_contact = 'http://%s:%s' % (nowip, str(settings.getHTTPPort()))
        if id_obj.getContactIndex(contact=http_contact) < 0:
            if _Debug:
                lg.out(4, 'http_interface.verify_contacts returning False: http contact not found or changed')
            return False
#         if http_node.get_internal_port() != settings.getHTTPPort():
#             if _Debug:
#                 lg.out(4, 'http_interface.verify_contacts returning False: http port has been changed')
#             return False
        if _Debug:
            lg.out(4, 'http_interface.verify_contacts returning True')
        return True

    def send_file(self, remote_idurl, filename, host, description='', keep_alive=True):
        """
        """
        return http_node.send_file(remote_idurl, filename)

    def send_file_single(self, remote_idurl, filename, host, description='', keep_alive=False):
        """
        """
        return http_node.send_file(remote_idurl, filename)


#     def connect_to(self, host):
#         """
#         """
#         return http_node.connect_to(host)

#     def disconnect_from(self, host):
#         """
#         """
#         return http_node.disconnect_from(host)

#     def cancel_file_sending(self, transferID):
#         """
#         """
#         return http_node.cancel_file_sending(transferID)

#     def cancel_file_receiving(self, transferID):
#         """
#         """
#         return http_node.cancel_file_receiving(transferID)

#     def cancel_outbox_file(self, host, filename):
#         """
#         """
#         return http_node.cancel_outbox_file(host, filename)

#     def list_sessions(self):
#         """
#         """
#         result = []
#         for opened_connection in http_node.opened_connections().values():
#             for channel in opened_connection:
#                 result.append(channel)
#         for started_connection in http_node.started_connections():
#             result.append(started_connection)
#         return result

#     def list_streams(self, sorted_by_time=True):
#         """
#         """
#         result = []
#         result.extend(http_node.list_input_streams(sorted_by_time))
#         result.extend(http_node.list_output_streams(sorted_by_time))
#         return result

#------------------------------------------------------------------------------


def interface_transport_initialized(xmlrpcurl):
    if proxy():
        return proxy().callRemote('transport_initialized', 'http', xmlrpcurl)
    lg.warn('transport_http is not ready')
    return fail(Exception('transport_http is not ready'))


def interface_receiving_started(host, new_options={}):
    if proxy():
        return proxy().callRemote('receiving_started', 'http', net_misc.pack_address(host), new_options)
    lg.warn('transport_http is not ready')
    return fail(Exception('transport_http is not ready'))


def interface_receiving_failed(error_code=None):
    if proxy():
        return proxy().callRemote('receiving_failed', 'http', error_code)
    lg.warn('transport_http is not ready')
    return fail(Exception('transport_http is not ready'))


def interface_disconnected(result=None):
    if proxy():
        return proxy().callRemote('disconnected', 'http', result)
    lg.warn('transport_http is not ready')
    return fail(Exception('transport_http is not ready'))


def interface_register_file_sending(host, receiver_idurl, filename, size=0, description=''):
    if proxy():
        return proxy().callRemote('register_file_sending', 'http', net_misc.pack_address(host), receiver_idurl, filename, size, description)
    lg.warn('transport_http is not ready')
    return fail(Exception('transport_http is not ready'))


def interface_register_file_receiving(host, sender_idurl, filename, size=0):
    if proxy():
        return proxy().callRemote('register_file_receiving', 'http', net_misc.pack_address(host), sender_idurl, filename, size)
    lg.warn('transport_http is not ready')
    return fail(Exception('transport_http is not ready'))


def interface_unregister_file_sending(transfer_id, status, size=0, error_message=None):
    if proxy():
        return proxy().callRemote('unregister_file_sending', transfer_id, status, size, error_message)
    lg.warn('transport_http is not ready')
    return fail(Exception('transport_http is not ready'))


def interface_unregister_file_receiving(transfer_id, status, size=0, error_message=None):
    if proxy():
        return proxy().callRemote('unregister_file_receiving', transfer_id, status, size, error_message)
    lg.warn('transport_http is not ready')
    return fail(Exception('transport_http is not ready'))


def interface_cancelled_file_sending(host, filename, size=0, description=None, error_message=None):
    if proxy():
        return proxy().callRemote('cancelled_file_sending', 'http', net_misc.pack_address(host), filename, size, description, error_message)
    lg.warn('transport_http is not ready')
    return fail(Exception('transport_http is not ready'))



#!/usr/bin/env python
# serialization.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (serialization.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

from bitdust.lib import jsn
from bitdust.lib import strng

#------------------------------------------------------------------------------


def DictToBytes(dct, encoding='latin1', errors='strict', keys_to_text=False, values_to_text=False, pack_types=False, indent=None):
    """
    Calls `json.dupms()` method for input dict to build bytes output.
    Uses encoding to decode every byte string to text and ensure ascii output.
    If `keys_to_text` is True it will also convert dict keys from binary strings to text strings.
    If `values_to_text` is True it will also convert dict values from binary strings to text strings.
    Another smart feature is `pack_types` - it will remember types of keys and values in the input dict.
    Respective feature `unpack_types` of `BytesToDict()` method can be used to "extract" exactly same dict from bytes.
    Can be used to serialize dictionaries of mixed types - with binary and text values.
    """

    result = strng.to_bin(
        jsn.dumps(
            jsn.pack_dict(dct, encoding=encoding, errors=errors) if pack_types else dct,
            indent=indent,
            separators=(',', ':'),
            sort_keys=True,
            ensure_ascii=True,
            encoding=encoding,
            keys_to_text=keys_to_text,
            values_to_text=values_to_text,
        ),
        encoding=encoding,
        errors=errors,
    )
    return result


def BytesToDict(inp, encoding='latin1', errors='strict', keys_to_text=False, values_to_text=False, unpack_types=False):
    """
    A smart way to extract input bytes into python dictionary object.
    All input bytes will be decoded into text and then loaded via `json.loads()` method.
    Finally every text key and value in result dict will be encoded back to bytes if `values_to_text` is False.
    Smart feature `unpack_types` can be used to "extract" real types of keys and values from input bytes.
    Can be used to extract dictionaries of mixed types - binary and text values.
    """
    if not inp:
        return {}
    _t = strng.to_text(inp, encoding=encoding)
    if values_to_text:
        return jsn.loads_text(_t, encoding=encoding)
    if unpack_types:
        return jsn.unpack_dict(jsn.loads(_t, encoding=encoding), encoding=encoding, errors=errors)
    if keys_to_text:
        return jsn.dict_keys_to_text(jsn.loads(_t, encoding=encoding))
    return jsn.loads(_t, encoding=encoding)



#!/usr/bin/env python
# datastore.py
#
# Copyright (C) 2007-2008 Francois Aucamp, Meraka Institute, CSIR
# See AUTHORS for all authors and contact information.
#
# License: GNU Lesser General Public License, version 3 or later; see COPYING
#          included in this archive for details.
#
# This library is free software, distributed under the terms of
# the GNU Lesser General Public License Version 3, or any later version.
# See the COPYING file included in this archive
#
# The docstrings in this module contain epytext markup; API documentation
# may be created by processing this file with epydoc: http://epydoc.sf.net

from __future__ import absolute_import

try:
    from UserDict import DictMixin
except ImportError:
    try:
        from collections.abc import MutableMapping as DictMixin
    except:
        from collections import MutableMapping as DictMixin

import sqlite3
import os
import json

from . import constants  # @UnresolvedImport
from . import encoding  # @UnresolvedImport

try:
    buffer = buffer  # @UndefinedVariable
except:
    buffer = memoryview

PROTOCOL_VERSION = 1

PICKLE_PROTOCOL = 2

_Debug = False


class DataStore(DictMixin):
    """
    Interface for classes implementing physical storage (for data published via
    the "STORE" RPC) for the Kademlia DHT.

    @note: This provides an interface for a dict-like object
    """
    def keys(self):
        """
        Return a list of the keys in this data store.
        """

    def lastPublished(self, key):
        """
        Get the time the C{(key, value)} pair identified by C{key} was last
        published.
        """

    def originalPublisherID(self, key):
        """
        Get the original publisher of the data's node ID.

        @param key: The key that identifies the stored data
        @type key: str

        @return: Return the node ID of the original publisher of the
        C{(key, value)} pair identified by C{key}.
        """

    def originalPublishTime(self, key):
        """
        Get the time the C{(key, value)} pair identified by C{key} was
        originally published.
        """

    def getItem(self, key):
        """
        """

    def setItem(self, key, value, lastPublished, originallyPublished, originalPublisherID, **kwargs):
        """
        Set the value of the (key, value) pair identified by C{key}; this
        should set the "last published" value for the (key, value) pair to the
        current time.
        """

    def __getitem__(self, key):
        """
        Get the value identified by C{key}
        """

    def __setitem__(self, key, value):
        """
        Convenience wrapper to C{setItem}; this accepts a tuple in the format:
        (value, lastPublished, originallyPublished, originalPublisherID)
        """
        self.setItem(key, *value)

    def __delitem__(self, key):
        """
        Delete the specified key (and its value)
        """

    def __iter__(self):
        """
        """
        return self

    def __next__(self):
        """
        """

    def __len__(self):
        """
        """


class DictDataStore(DataStore):
    """
    A datastore using an in-memory Python dictionary.
    """
    def __init__(self):
        # Dictionary format:
        # { <key>: (<value>, <lastPublished>, <originallyPublished> <originalPublisherID>) }
        self._dict = {}

    def keys(self):
        """
        Return a list of the keys in this data store.
        """
        return list(self._dict.keys())

    def lastPublished(self, key):
        """
        Get the time the C{(key, value)} pair identified by C{key} was last
        published.
        """
        return self._dict[key][1]

    def originalPublisherID(self, key):
        """
        Get the original publisher of the data's node ID.

        @param key: The key that identifies the stored data
        @type key: str

        @return: Return the node ID of the original publisher of the
        C{(key, value)} pair identified by C{key}.
        """
        return self._dict[key][3]

    def originalPublishTime(self, key):
        """
        Get the time the C{(key, value)} pair identified by C{key} was
        originally published.
        """
        return self._dict[key][2]

    def setItem(self, key, value, lastPublished, originallyPublished, originalPublisherID, **kwargs):
        """
        Set the value of the (key, value) pair identified by C{key}; this
        should set the "last published" value for the (key, value) pair to the
        current time.
        """
        self._dict[key] = (value, lastPublished, originallyPublished, originalPublisherID)

    def __getitem__(self, key):
        """
        Get the value identified by C{key}
        """
        return self._dict[key][0]

    def __delitem__(self, key):
        """
        Delete the specified key (and its value)
        """
        del self._dict[key]

    def getItem(self, key):
        try:
            row = self._dict[key]
            result = dict(
                key=row[0].encode(),
                value=row[1],
                lastPublished=row[2],
                originallyPublished=row[3],
                originalPublisherID=None if not row[4] else row[4].encode(),
            )
        except:
            return None
        return result


class SQLiteVersionedJsonDataStore(DataStore):
    """
    SQLite database-based datastore.
    """
    def __init__(self, dbFile=':memory:'):
        """
        @param dbFile: The name of the file containing the SQLite database; if
                       unspecified, an in-memory database is used.
        @type dbFile: str
        """
        self.dbFile = dbFile
        createDB = not os.path.exists(dbFile)
        if _Debug:
            print('[DHT DB] dbFile=%r   createDB=%r' % (
                dbFile,
                createDB,
            ))
        self._db = sqlite3.connect(dbFile)
        self._db.isolation_level = None
        self._db.text_factory = encoding.to_text
        if createDB:
            self.create_table()
            if _Debug:
                print('[DHT DB]  Created empty table for DHT records')
        self._cursor = self._db.cursor()

    def _dbQuery(self, key, columnName):
        try:
            self._cursor.execute(
                'SELECT %s FROM data WHERE key=:reqKey' % columnName,
                {
                    'reqKey': key,
                },
            )
            row = self._cursor.fetchone()
            value = row[0]
        except:
            raise KeyError(key)
        else:
            return value

    def __getitem__(self, key):
        v = self._dbQuery(key, 'value')
        v = json.loads(v)
        return v['d']

    def __delitem__(self, key):
        self._cursor.execute('DELETE FROM data WHERE key=:reqKey', {
            'reqKey': key,
        })

    def create_table(self):
        self._db.execute('CREATE TABLE data(key, value, lastPublished, originallyPublished, originalPublisherID, expireSeconds, revision)')

    def keys(self):
        """
        Return a list of the keys in this data store.
        """
        keys = []
        try:
            self._cursor.execute('SELECT key FROM data')
            for row in self._cursor:
                keys.append(row[0])
        finally:
            return keys

    def lastPublished(self, key):
        """
        Get the time the C{(key, value)} pair identified by C{key} was last
        published.
        """
        return int(self._dbQuery(key, 'lastPublished'))

    def originalPublisherID(self, key):
        """
        Get the original publisher of the data's node ID.

        @param key: The key that identifies the stored data
        @type key: str

        @return: Return the node ID of the original publisher of the
        C{(key, value)} pair identified by C{key}.
        """
        return self._dbQuery(key, 'originalPublisherID')

    def originalPublishTime(self, key):
        """
        Get the time the C{(key, value)} pair identified by C{key} was
        originally published.
        """
        return int(self._dbQuery(key, 'originallyPublished'))

    def expireSeconds(self, key):
        """
        """
        return int(self._dbQuery(key, 'expireSeconds'))

    def revision(self, key):
        """
        """
        try:
            return int(self._dbQuery(key, 'revision'))
        except KeyError:
            return 0

    def setItem(self, key, value, lastPublished, originallyPublished, originalPublisherID, expireSeconds=constants.dataExpireSecondsDefaut, **kwargs):
        key_hex = encoding.to_text(key)
        new_revision = kwargs.get('revision', None)
        if new_revision is None:
            new_revision = self.revision(key) + 1
        self._cursor.execute('select key from data where key=:reqKey', {'reqKey': key_hex})
        opID = originalPublisherID or None
        if self._cursor.fetchone() is None:
            self._cursor.execute(
                'INSERT INTO data(key, value, lastPublished, originallyPublished, originalPublisherID, expireSeconds, revision) VALUES (?, ?, ?, ?, ?, ?, ?)', (
                    key_hex,
                    json.dumps({
                        'k': key_hex,
                        'd': value,
                        'v': PROTOCOL_VERSION,
                    }),
                    lastPublished,
                    originallyPublished,
                    opID,
                    expireSeconds,
                    new_revision,
                )
            )
            if _Debug:
                print('[DHT DB] %r setItem  stored new value for key [%s] with revision %d' % (self.dbFile, key, new_revision))
        else:
            self._cursor.execute(
                'UPDATE data SET value=?, lastPublished=?, originallyPublished=?, originalPublisherID=?, expireSeconds=?, revision=? WHERE key=?', (
                    json.dumps({
                        'k': key_hex,
                        'd': value,
                        'v': PROTOCOL_VERSION,
                    }),
                    lastPublished,
                    originallyPublished,
                    opID,
                    expireSeconds,
                    new_revision,
                    key_hex,
                )
            )
            if _Debug:
                print('[DHT DB] %r setItem  updated existing value for key [%s] with revision %d' % (self.dbFile, key, new_revision))

    def getItem(self, key):
        key_hex = key
        key_hex = encoding.to_text(key)
        self._cursor.execute('SELECT * FROM data WHERE key=:reqKey', {
            'reqKey': key_hex,
        })

        row = self._cursor.fetchone()
        if not row:
            if _Debug:
                print('[DHT DB] %r getItem [%s]  return None : did not found key in dataStore' % (self.dbfile, key))
            return None

        v = row[1]
        if isinstance(v, buffer):
            v = encoding.to_text(v)

        v = json.loads(v)

        # TODO: check / verify v['k'] against key_hex
        # TODO: check / verify v['v'] against PROTOCOL_VERSION

        value = v['d']

        key_orig = row[0]

        # TODO: check / verify key_orig against key

        opID = row[4] or None

        result = dict(
            key=key_orig,
            value=value,
            lastPublished=row[2],
            originallyPublished=row[3],
            originalPublisherID=opID,
            expireSeconds=row[5],
            revision=row[6],
        )

        if _Debug:
            print('[DHT DB] %r getItem   found one record for key [%s], revision is %d' % (self.dbFile, key, row[6]))
        return result

    def getAllItems(self):
        self._cursor.execute('SELECT * FROM data')
        rows = self._cursor.fetchall()
        items = []
        for row in rows:

            v = row[1]
            if isinstance(v, buffer):
                v = encoding.to_text(v)

            v = json.loads(v)

            # TODO: check / verify v['k'] against key_hex
            # TODO: check / verify v['v'] against PROTOCOL_VERSION

            value = v['d']

            _k = row[0]
            _opID = row[4] or None

            items.append(dict(
                value=value,
                lastPublished=row[2],
                originallyPublished=row[3],
                originalPublisherID=_opID,
                expireSeconds=row[5],
                revision=row[6],
            ))
        return items



import logging, sys
from logging.handlers import RotatingFileHandler


def filter_status(record):
    """"
    Only displays log messages about status info
    or ERROR level
    """
    if ('Status:' in str(record.msg)) or (record.levelname == 'ERROR'):
        return 1
    else:
        return 0


def log(logFile, level_input='WARNING', terminal_output=False):
    if level_input == 'NOTSET':
        level = logging.NOTSET
    if level_input == 'DEBUG':
        level = logging.DEBUG
    if level_input == 'INFO':
        level = logging.INFO
    if level_input == 'WARNING':
        level = logging.WARNING
    if level_input == 'ERROR':
        level = logging.ERROR
    if level_input == 'CRITICAL':
        level = logging.CRITICAL

    log_formatter = logging.Formatter('%(asctime)s %(levelname)s %(funcName)s(%(lineno)d) %(message)s')
    my_handler = RotatingFileHandler(logFile, mode='a', maxBytes=5*1024*1024, backupCount=2, encoding='utf-8', delay=0)
    my_handler.setFormatter(log_formatter)
    my_handler.setLevel(level)
    app_log = logging.getLogger('root')
    app_log.setLevel(level)
    app_log.addHandler(my_handler)

    # This part is what goes on console.
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(level)
    # TODO: We could have 2 level in the config, one for screen and one for files.
    # print('Logging level: {} ({})'.format(level_input, level))
    if terminal_output != True:
        ch.addFilter(filter_status)
        # No need for complete func and line info here.
        formatter = logging.Formatter('%(asctime)s %(message)s')
    else:
        formatter = logging.Formatter('%(asctime)s %(funcName)s(%(lineno)d) %(message)s')
    ch.setFormatter(formatter)
    app_log.addHandler(ch)

    return app_log
